from __future__ import print_function, unicode_literals
from zope.interface import alsoProvides
from twisted.trial import unittest
from twisted.internet.task import Clock, Cooperator
from twisted.internet.interfaces import IStreamServerEndpoint
from unittest import mock
from ...eventual import EventualQueue
from ..._interfaces import ISend, ITerminator, ISubChannel
from ...util import dict_to_bytes
from ..._dilation import roles
from ..._dilation.manager import (Dilator, Manager, make_side,
                                  OldPeerCannotDilateError,
                                  UnknownDilationMessageType,
                                  UnexpectedKCM,
                                  UnknownMessageType)
from ..._dilation.connection import Open, Data, Close, Ack, KCM, Ping, Pong
from ..._dilation.subchannel import _SubchannelAddress
from .common import clear_mock_calls


class Holder():
    pass


def make_dilator():
    h = Holder()
    h.reactor = object()
    h.clock = Clock()
    h.eq = EventualQueue(h.clock)
    term = mock.Mock(side_effect=lambda: True)  # one write per Eventual tick

    def term_factory():
        return term
    h.coop = Cooperator(terminationPredicateFactory=term_factory,
                        scheduler=h.eq.eventually)
    h.send = mock.Mock()
    alsoProvides(h.send, ISend)
    dil = Dilator(h.reactor, h.eq, h.coop)
    h.terminator = mock.Mock()
    alsoProvides(h.terminator, ITerminator)
    dil.wire(h.send, h.terminator)
    return dil, h


class TestDilator(unittest.TestCase):
    # we should test the interleavings between:
    # * application calls w.dilate() and gets back endpoints
    # * wormhole gets: dilation key, VERSION, 0-n dilation messages

    def test_dilate_first(self):
        (dil, h) = make_dilator()
        side = object()
        m = mock.Mock()
        eps = object()
        m.get_endpoints = mock.Mock(return_value=eps)
        mm = mock.Mock(side_effect=[m])
        with mock.patch("wormhole._dilation.manager.Manager", mm), \
             mock.patch("wormhole._dilation.manager.make_side",
                        return_value=side):
            eps1 = dil.dilate()
            eps2 = dil.dilate()
        self.assertIdentical(eps1, eps)
        self.assertIdentical(eps1, eps2)
        self.assertEqual(mm.mock_calls, [mock.call(h.send, side, None,
                                                   h.reactor, h.eq, h.coop, False)])

        self.assertEqual(m.mock_calls, [mock.call.get_endpoints(),
                                        mock.call.get_endpoints()])
        clear_mock_calls(m)

        key = b"key"
        transit_key = object()
        with mock.patch("wormhole._dilation.manager.derive_key",
                        return_value=transit_key) as dk:
            dil.got_key(key)
        self.assertEqual(dk.mock_calls, [mock.call(key, b"dilation-v1", 32)])
        self.assertEqual(m.mock_calls, [mock.call.got_dilation_key(transit_key)])
        clear_mock_calls(m)

        wv = object()
        dil.got_wormhole_versions(wv)
        self.assertEqual(m.mock_calls, [mock.call.got_wormhole_versions(wv)])
        clear_mock_calls(m)

        dm1 = object()
        dm2 = object()
        dil.received_dilate(dm1)
        dil.received_dilate(dm2)
        self.assertEqual(m.mock_calls, [mock.call.received_dilation_message(dm1),
                                        mock.call.received_dilation_message(dm2),
                                        ])
        clear_mock_calls(m)

        stopped_d = mock.Mock()
        m.when_stopped = mock.Mock(return_value=stopped_d)
        dil.stop()
        self.assertEqual(m.mock_calls, [mock.call.stop(),
                                        mock.call.when_stopped(),
                                        ])

    def test_dilate_later(self):
        (dil, h) = make_dilator()
        m = mock.Mock()
        mm = mock.Mock(side_effect=[m])

        key = b"key"
        transit_key = object()
        with mock.patch("wormhole._dilation.manager.derive_key",
                        return_value=transit_key) as dk:
            dil.got_key(key)
        self.assertEqual(dk.mock_calls, [mock.call(key, b"dilation-v1", 32)])

        wv = object()
        dil.got_wormhole_versions(wv)

        dm1 = object()
        dil.received_dilate(dm1)

        self.assertEqual(mm.mock_calls, [])

        with mock.patch("wormhole._dilation.manager.Manager", mm):
            dil.dilate()
        self.assertEqual(m.mock_calls, [mock.call.got_dilation_key(transit_key),
                                        mock.call.got_wormhole_versions(wv),
                                        mock.call.received_dilation_message(dm1),
                                        mock.call.get_endpoints(),
                                        ])
        clear_mock_calls(m)

        dm2 = object()
        dil.received_dilate(dm2)
        self.assertEqual(m.mock_calls, [mock.call.received_dilation_message(dm2),
                                        ])

    def test_stop_early(self):
        (dil, h) = make_dilator()
        # we stop before w.dilate(), so there is no Manager to stop
        dil.stop()
        self.assertEqual(h.terminator.mock_calls, [mock.call.stoppedD()])

    def test_peer_cannot_dilate(self):
        (dil, h) = make_dilator()
        eps = dil.dilate()

        dil.got_key(b"\x01" * 32)
        dil.got_wormhole_versions({})  # missing "can-dilate"
        d = eps.connect.connect(None)
        h.eq.flush_sync()
        self.failureResultOf(d).check(OldPeerCannotDilateError)

    def test_disjoint_versions(self):
        (dil, h) = make_dilator()
        eps = dil.dilate()

        dil.got_key(b"\x01" * 32)
        dil.got_wormhole_versions({"can-dilate": ["-1"]})
        d = eps.connect.connect(None)
        h.eq.flush_sync()
        self.failureResultOf(d).check(OldPeerCannotDilateError)

    def test_transit_relay(self):
        (dil, h) = make_dilator()
        transit_relay_location = object()
        side = object()
        m = mock.Mock()
        mm = mock.Mock(side_effect=[m])
        with mock.patch("wormhole._dilation.manager.Manager", mm), \
             mock.patch("wormhole._dilation.manager.make_side",
                        return_value=side):
            dil.dilate(transit_relay_location)
        self.assertEqual(mm.mock_calls, [mock.call(h.send, side, transit_relay_location,
                                                   h.reactor, h.eq, h.coop, False)])


LEADER = "ff3456abcdef"
FOLLOWER = "123456abcdef"


def make_manager(leader=True):
    h = Holder()
    h.send = mock.Mock()
    alsoProvides(h.send, ISend)
    if leader:
        side = LEADER
    else:
        side = FOLLOWER
    h.key = b"\x00" * 32
    h.relay = None
    h.reactor = object()
    h.clock = Clock()
    h.eq = EventualQueue(h.clock)
    term = mock.Mock(side_effect=lambda: True)  # one write per Eventual tick

    def term_factory():
        return term
    h.coop = Cooperator(terminationPredicateFactory=term_factory,
                        scheduler=h.eq.eventually)
    h.inbound = mock.Mock()
    h.Inbound = mock.Mock(return_value=h.inbound)
    h.outbound = mock.Mock()
    h.Outbound = mock.Mock(return_value=h.outbound)
    h.sc0 = mock.Mock()
    alsoProvides(h.sc0, ISubChannel)
    h.SubChannel = mock.Mock(return_value=h.sc0)
    h.listen_ep = mock.Mock()
    alsoProvides(h.listen_ep, IStreamServerEndpoint)
    with mock.patch("wormhole._dilation.manager.Inbound", h.Inbound), \
         mock.patch("wormhole._dilation.manager.Outbound", h.Outbound), \
         mock.patch("wormhole._dilation.manager.SubChannel", h.SubChannel), \
         mock.patch("wormhole._dilation.manager.SubchannelListenerEndpoint",
                    return_value=h.listen_ep):
        m = Manager(h.send, side, h.relay, h.reactor, h.eq, h.coop)
    h.hostaddr = m._host_addr
    m.got_dilation_key(h.key)
    return m, h


class TestManager(unittest.TestCase):
    def test_make_side(self):
        side = make_side()
        self.assertEqual(type(side), type(u""))
        self.assertEqual(len(side), 2 * 8)

    def test_create(self):
        m, h = make_manager()

    def test_leader(self):
        m, h = make_manager(leader=True)
        self.assertEqual(h.send.mock_calls, [])
        self.assertEqual(h.Inbound.mock_calls, [mock.call(m, h.hostaddr)])
        self.assertEqual(h.Outbound.mock_calls, [mock.call(m, h.coop)])
        scid0 = 0
        sc0_peer_addr = _SubchannelAddress(scid0)
        self.assertEqual(h.SubChannel.mock_calls, [
            mock.call(scid0, m, m._host_addr, sc0_peer_addr),
            ])
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.set_subchannel_zero(scid0, h.sc0),
            mock.call.set_listener_endpoint(h.listen_ep)
            ])
        clear_mock_calls(h.inbound)

        eps = m.get_endpoints()
        self.assertTrue(hasattr(eps, "control"))
        self.assertTrue(hasattr(eps, "connect"))
        self.assertEqual(eps.listen, h.listen_ep)

        m.got_wormhole_versions({"can-dilate": ["1"]})
        self.assertEqual(h.send.mock_calls, [
            mock.call.send("dilate-0",
                           dict_to_bytes({"type": "please", "side": LEADER, "use-version": "1"}))
            ])
        clear_mock_calls(h.send)

        # ignore early hints
        m.rx_HINTS({})
        self.assertEqual(h.send.mock_calls, [])

        c = mock.Mock()
        connector = mock.Mock(return_value=c)
        with mock.patch("wormhole._dilation.manager.Connector", connector):
            # receiving this PLEASE triggers creation of the Connector
            m.rx_PLEASE({"side": FOLLOWER})
        self.assertEqual(h.send.mock_calls, [])
        self.assertEqual(connector.mock_calls, [
            mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                      False,  # no_listen
                      None,  # tor
                      None,  # timing
                      LEADER, roles.LEADER),
            ])
        self.assertEqual(c.mock_calls, [mock.call.start()])
        clear_mock_calls(connector, c)

        # now any inbound hints should get passed to our Connector
        with mock.patch("wormhole._dilation.manager.parse_hint",
                        side_effect=["p1", None, "p3"]) as ph:
            m.rx_HINTS({"hints": [1, 2, 3]})
        self.assertEqual(ph.mock_calls, [mock.call(1), mock.call(2), mock.call(3)])
        self.assertEqual(c.mock_calls, [mock.call.got_hints(["p1", "p3"])])
        clear_mock_calls(ph, c)

        # and we send out any (listening) hints from our Connector
        m.send_hints([1, 2])
        self.assertEqual(h.send.mock_calls, [
            mock.call.send("dilate-1",
                           dict_to_bytes({"type": "connection-hints",
                                          "hints": [1, 2]}))
            ])
        clear_mock_calls(h.send)

        # the first successful connection fires when_first_connected(), so
        # the endpoints can activate
        c1 = mock.Mock()
        m.connector_connection_made(c1)

        self.assertEqual(h.inbound.mock_calls, [mock.call.use_connection(c1)])
        self.assertEqual(h.outbound.mock_calls, [mock.call.use_connection(c1)])
        clear_mock_calls(h.inbound, h.outbound)

        # the Leader making a new outbound channel should get scid=1
        scid1 = 1
        self.assertEqual(m.allocate_subchannel_id(), scid1)
        r1 = Open(10, scid1)  # seqnum=10
        h.outbound.build_record = mock.Mock(return_value=r1)
        m.send_open(scid1)
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.build_record(Open, scid1),
            mock.call.queue_and_send_record(r1),
            ])
        clear_mock_calls(h.outbound)

        r2 = Data(11, scid1, b"data")
        h.outbound.build_record = mock.Mock(return_value=r2)
        m.send_data(scid1, b"data")
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.build_record(Data, scid1, b"data"),
            mock.call.queue_and_send_record(r2),
            ])
        clear_mock_calls(h.outbound)

        r3 = Close(12, scid1)
        h.outbound.build_record = mock.Mock(return_value=r3)
        m.send_close(scid1)
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.build_record(Close, scid1),
            mock.call.queue_and_send_record(r3),
            ])
        clear_mock_calls(h.outbound)

        # ack the OPEN
        m.got_record(Ack(10))
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.handle_ack(10)
            ])
        clear_mock_calls(h.outbound)

        # test that inbound records get acked and routed to Inbound
        h.inbound.is_record_old = mock.Mock(return_value=False)
        scid2 = 2
        o200 = Open(200, scid2)
        m.got_record(o200)
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.send_if_connected(Ack(200))
            ])
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.is_record_old(o200),
            mock.call.update_ack_watermark(200),
            mock.call.handle_open(scid2),
            ])
        clear_mock_calls(h.outbound, h.inbound)

        # old (duplicate) records should provoke new Acks, but not get
        # forwarded
        h.inbound.is_record_old = mock.Mock(return_value=True)
        m.got_record(o200)
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.send_if_connected(Ack(200))
            ])
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.is_record_old(o200),
            ])
        clear_mock_calls(h.outbound, h.inbound)

        # check Data and Close too
        h.inbound.is_record_old = mock.Mock(return_value=False)
        d201 = Data(201, scid2, b"data")
        m.got_record(d201)
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.send_if_connected(Ack(201))
            ])
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.is_record_old(d201),
            mock.call.update_ack_watermark(201),
            mock.call.handle_data(scid2, b"data"),
            ])
        clear_mock_calls(h.outbound, h.inbound)

        c202 = Close(202, scid2)
        m.got_record(c202)
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.send_if_connected(Ack(202))
            ])
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.is_record_old(c202),
            mock.call.update_ack_watermark(202),
            mock.call.handle_close(scid2),
            ])
        clear_mock_calls(h.outbound, h.inbound)

        # Now we lose the connection. The Leader should tell the other side
        # that we're reconnecting.

        m.connector_connection_lost()
        self.assertEqual(h.send.mock_calls, [
            mock.call.send("dilate-2",
                           dict_to_bytes({"type": "reconnect"}))
            ])
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.stop_using_connection()
            ])
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.stop_using_connection()
            ])
        clear_mock_calls(h.send, h.inbound, h.outbound)

        # leader does nothing (stays in FLUSHING) until the follower acks by
        # sending RECONNECTING

        # inbound hints should be ignored during FLUSHING
        with mock.patch("wormhole._dilation.manager.parse_hint",
                        return_value=None) as ph:
            m.rx_HINTS({"hints": [1, 2, 3]})
        self.assertEqual(ph.mock_calls, [])  # ignored

        c2 = mock.Mock()
        connector2 = mock.Mock(return_value=c2)
        with mock.patch("wormhole._dilation.manager.Connector", connector2):
            # this triggers creation of a new Connector
            m.rx_RECONNECTING()
        self.assertEqual(h.send.mock_calls, [])
        self.assertEqual(connector2.mock_calls, [
            mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                      False,  # no_listen
                      None,  # tor
                      None,  # timing
                      LEADER, roles.LEADER),
            ])
        self.assertEqual(c2.mock_calls, [mock.call.start()])
        clear_mock_calls(connector2, c2)

        self.assertEqual(h.inbound.mock_calls, [])
        self.assertEqual(h.outbound.mock_calls, [])

        # and a new connection should re-register with Inbound/Outbound,
        # which are responsible for re-sending unacked queued messages
        c3 = mock.Mock()
        m.connector_connection_made(c3)

        self.assertEqual(h.inbound.mock_calls, [mock.call.use_connection(c3)])
        self.assertEqual(h.outbound.mock_calls, [mock.call.use_connection(c3)])
        clear_mock_calls(h.inbound, h.outbound)

    def test_follower(self):
        m, h = make_manager(leader=False)

        m.got_wormhole_versions({"can-dilate": ["1"]})
        self.assertEqual(h.send.mock_calls, [
            mock.call.send("dilate-0",
                           dict_to_bytes({"type": "please", "side": FOLLOWER, "use-version": "1"}))
            ])
        clear_mock_calls(h.send)
        clear_mock_calls(h.inbound)

        c = mock.Mock()
        connector = mock.Mock(return_value=c)
        with mock.patch("wormhole._dilation.manager.Connector", connector):
            # receiving this PLEASE triggers creation of the Connector
            m.rx_PLEASE({"side": LEADER})
        self.assertEqual(h.send.mock_calls, [])
        self.assertEqual(connector.mock_calls, [
            mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                      False,  # no_listen
                      None,  # tor
                      None,  # timing
                      FOLLOWER, roles.FOLLOWER),
            ])
        self.assertEqual(c.mock_calls, [mock.call.start()])
        clear_mock_calls(connector, c)

        # get connected, then lose the connection
        c1 = mock.Mock()
        m.connector_connection_made(c1)
        self.assertEqual(h.inbound.mock_calls, [mock.call.use_connection(c1)])
        self.assertEqual(h.outbound.mock_calls, [mock.call.use_connection(c1)])
        clear_mock_calls(h.inbound, h.outbound)

        # now lose the connection. As the follower, we don't notify the
        # leader, we just wait for them to notice
        m.connector_connection_lost()
        self.assertEqual(h.send.mock_calls, [])
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.stop_using_connection()
            ])
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.stop_using_connection()
            ])
        clear_mock_calls(h.send, h.inbound, h.outbound)

        # now we get a RECONNECT: we should send RECONNECTING
        c2 = mock.Mock()
        connector2 = mock.Mock(return_value=c2)
        with mock.patch("wormhole._dilation.manager.Connector", connector2):
            m.rx_RECONNECT()
        self.assertEqual(h.send.mock_calls, [
            mock.call.send("dilate-1",
                           dict_to_bytes({"type": "reconnecting"}))
            ])
        self.assertEqual(connector2.mock_calls, [
            mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                      False,  # no_listen
                      None,  # tor
                      None,  # timing
                      FOLLOWER, roles.FOLLOWER),
            ])
        self.assertEqual(c2.mock_calls, [mock.call.start()])
        clear_mock_calls(connector2, c2)

        # while we're trying to connect, we get told to stop again, so we
        # should abandon the connection attempt and start another
        c3 = mock.Mock()
        connector3 = mock.Mock(return_value=c3)
        with mock.patch("wormhole._dilation.manager.Connector", connector3):
            m.rx_RECONNECT()
        self.assertEqual(c2.mock_calls, [mock.call.stop()])
        self.assertEqual(connector3.mock_calls, [
            mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                      False,  # no_listen
                      None,  # tor
                      None,  # timing
                      FOLLOWER, roles.FOLLOWER),
            ])
        self.assertEqual(c3.mock_calls, [mock.call.start()])
        clear_mock_calls(c2, connector3, c3)

        m.connector_connection_made(c3)
        # finally if we're already connected, rx_RECONNECT means we should
        # abandon this connection (even though it still looks ok to us), then
        # when the attempt is finished stopping, we should start another

        m.rx_RECONNECT()

        c4 = mock.Mock()
        connector4 = mock.Mock(return_value=c4)
        with mock.patch("wormhole._dilation.manager.Connector", connector4):
            m.connector_connection_lost()
        self.assertEqual(c3.mock_calls, [mock.call.disconnect()])
        self.assertEqual(connector4.mock_calls, [
            mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                      False,  # no_listen
                      None,  # tor
                      None,  # timing
                      FOLLOWER, roles.FOLLOWER),
            ])
        self.assertEqual(c4.mock_calls, [mock.call.start()])
        clear_mock_calls(c3, connector4, c4)

    def test_mirror(self):
        # receive a PLEASE with the same side as us: shouldn't happen
        m, h = make_manager(leader=True)

        m.start()
        clear_mock_calls(h.send)
        e = self.assertRaises(ValueError, m.rx_PLEASE, {"side": LEADER})
        self.assertEqual(str(e), "their side shouldn't be equal: reflection?")

    def test_ping_pong(self):
        m, h = make_manager(leader=False)

        m.got_record(KCM())
        self.flushLoggedErrors(UnexpectedKCM)

        m.got_record(Ping(1))
        self.assertEqual(h.outbound.mock_calls,
                         [mock.call.send_if_connected(Pong(1))])
        clear_mock_calls(h.outbound)

        m.got_record(Pong(2))
        # currently ignored, will eventually update a timer

        m.got_record("not recognized")
        e = self.flushLoggedErrors(UnknownMessageType)
        self.assertEqual(len(e), 1)
        self.assertEqual(str(e[0].value), "not recognized")

        m.send_ping(3)
        self.assertEqual(h.outbound.mock_calls,
                         [mock.call.send_if_connected(Pong(3))])
        clear_mock_calls(h.outbound)

    def test_subchannel(self):
        m, h = make_manager(leader=True)
        clear_mock_calls(h.inbound)
        sc = object()

        m.subchannel_pauseProducing(sc)
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.subchannel_pauseProducing(sc)])
        clear_mock_calls(h.inbound)

        m.subchannel_resumeProducing(sc)
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.subchannel_resumeProducing(sc)])
        clear_mock_calls(h.inbound)

        m.subchannel_stopProducing(sc)
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.subchannel_stopProducing(sc)])
        clear_mock_calls(h.inbound)

        p = object()
        streaming = object()

        m.subchannel_registerProducer(sc, p, streaming)
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.subchannel_registerProducer(sc, p, streaming)])
        clear_mock_calls(h.outbound)

        m.subchannel_unregisterProducer(sc)
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.subchannel_unregisterProducer(sc)])
        clear_mock_calls(h.outbound)

        m.subchannel_closed(4, sc)
        self.assertEqual(h.inbound.mock_calls, [
            mock.call.subchannel_closed(4, sc)])
        self.assertEqual(h.outbound.mock_calls, [
            mock.call.subchannel_closed(4, sc)])
        clear_mock_calls(h.inbound, h.outbound)

    def test_unknown_message(self):
        # receive a PLEASE with the same side as us: shouldn't happen
        m, h = make_manager(leader=True)
        m.start()

        m.received_dilation_message(dict_to_bytes(dict(type="unknown")))
        self.flushLoggedErrors(UnknownDilationMessageType)

    # TODO: test transit relay is used



from __future__ import print_function, unicode_literals

from twisted.internet import reactor
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.task import Clock
from twisted.trial import unittest

from unittest import mock

from ..eventual import EventualQueue


class IntentionalError(Exception):
    pass


class Eventual(unittest.TestCase, object):
    def test_eventually(self):
        c = Clock()
        eq = EventualQueue(c)
        c1 = mock.Mock()
        eq.eventually(c1, "arg1", "arg2", kwarg1="kw1")
        eq.eventually(c1, "arg3", "arg4", kwarg5="kw5")
        d2 = eq.fire_eventually()
        d3 = eq.fire_eventually("value")
        self.assertEqual(c1.mock_calls, [])
        self.assertNoResult(d2)
        self.assertNoResult(d3)

        eq.flush_sync()
        self.assertEqual(c1.mock_calls, [
            mock.call("arg1", "arg2", kwarg1="kw1"),
            mock.call("arg3", "arg4", kwarg5="kw5")
        ])
        self.assertEqual(self.successResultOf(d2), None)
        self.assertEqual(self.successResultOf(d3), "value")

    def test_error(self):
        c = Clock()
        eq = EventualQueue(c)
        c1 = mock.Mock(side_effect=IntentionalError)
        eq.eventually(c1, "arg1", "arg2", kwarg1="kw1")
        self.assertEqual(c1.mock_calls, [])

        eq.flush_sync()
        self.assertEqual(c1.mock_calls,
                         [mock.call("arg1", "arg2", kwarg1="kw1")])

        self.flushLoggedErrors(IntentionalError)

    @inlineCallbacks
    def test_flush(self):
        eq = EventualQueue(reactor)
        d1 = eq.fire_eventually()
        d2 = Deferred()

        def _more(res):
            eq.eventually(d2.callback, None)

        d1.addCallback(_more)
        yield eq.flush()
        # d1 will fire, which will queue d2 to fire, and the flush() ought to
        # wait for d2 too
        self.successResultOf(d2)



from __future__ import print_function, unicode_literals
from unittest import mock
from zope.interface import alsoProvides
from twisted.trial import unittest
from twisted.internet.interfaces import ITransport
from ..._dilation._noise import NoiseInvalidMessage
from ..._dilation.connection import (IFramer, Frame, Prologue,
                                     _Record, Handshake, KCM,
                                     Disconnect, Ping, _Framer, Data)
from ..._dilation.connector import build_noise
from ..._dilation.roles import LEADER, FOLLOWER
from zope.interface import implementer


def make_record():
    f = mock.Mock()
    alsoProvides(f, IFramer)
    n = mock.Mock()  # pretends to be a Noise object
    r = _Record(f, n, LEADER)
    r.set_role_leader()
    return r, f, n


class Record(unittest.TestCase):
    def test_good2(self):
        f = mock.Mock()
        alsoProvides(f, IFramer)
        f.add_and_parse = mock.Mock(side_effect=[
            [],
            [Prologue()],
            [Frame(frame=b"rx-handshake")],
            [Frame(frame=b"frame1"), Frame(frame=b"frame2")],
        ])
        n = mock.Mock()
        n.write_message = mock.Mock(return_value=b"tx-handshake")
        p1, p2 = object(), object()
        n.decrypt = mock.Mock(side_effect=[p1, p2])
        r = _Record(f, n, LEADER)
        r.set_role_leader()
        self.assertEqual(f.mock_calls, [])
        r.connectionMade()
        self.assertEqual(f.mock_calls, [mock.call.connectionMade()])
        f.mock_calls[:] = []
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        n.mock_calls[:] = []

        # Pretend to deliver the prologue in two parts. The text we send in
        # doesn't matter: the side_effect= is what causes the prologue to be
        # recognized by the second call.
        self.assertEqual(list(r.add_and_unframe(b"pro")), [])
        self.assertEqual(f.mock_calls, [mock.call.add_and_parse(b"pro")])
        f.mock_calls[:] = []
        self.assertEqual(n.mock_calls, [])

        # recognizing the prologue causes a handshake frame to be sent
        self.assertEqual(list(r.add_and_unframe(b"logue")), [])
        self.assertEqual(f.mock_calls, [mock.call.add_and_parse(b"logue"),
                                        mock.call.send_frame(b"tx-handshake")])
        f.mock_calls[:] = []
        self.assertEqual(n.mock_calls, [mock.call.write_message()])
        n.mock_calls[:] = []

        # next add_and_unframe is recognized as the Handshake
        self.assertEqual(list(r.add_and_unframe(b"blah")), [Handshake()])
        self.assertEqual(f.mock_calls, [mock.call.add_and_parse(b"blah")])
        f.mock_calls[:] = []
        self.assertEqual(n.mock_calls, [mock.call.read_message(b"rx-handshake")])
        n.mock_calls[:] = []

        # next is a pair of Records
        r1, r2 = object(), object()
        with mock.patch("wormhole._dilation.connection.parse_record",
                        side_effect=[r1, r2]) as pr:
            self.assertEqual(list(r.add_and_unframe(b"blah2")), [r1, r2])
            self.assertEqual(n.mock_calls, [mock.call.decrypt(b"frame1"),
                                            mock.call.decrypt(b"frame2")])
            self.assertEqual(pr.mock_calls, [mock.call(p1), mock.call(p2)])

    def test_bad_handshake(self):
        f = mock.Mock()
        alsoProvides(f, IFramer)
        f.add_and_parse = mock.Mock(return_value=[Prologue(),
                                                  Frame(frame=b"rx-handshake")])
        n = mock.Mock()
        n.write_message = mock.Mock(return_value=b"tx-handshake")
        nvm = NoiseInvalidMessage()
        n.read_message = mock.Mock(side_effect=nvm)
        r = _Record(f, n, LEADER)
        r.set_role_leader()
        self.assertEqual(f.mock_calls, [])
        r.connectionMade()
        self.assertEqual(f.mock_calls, [mock.call.connectionMade()])
        f.mock_calls[:] = []
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        n.mock_calls[:] = []

        with mock.patch("wormhole._dilation.connection.log.err") as le:
            with self.assertRaises(Disconnect):
                list(r.add_and_unframe(b"data"))
        self.assertEqual(le.mock_calls,
                         [mock.call(nvm, "bad inbound noise handshake")])

    def test_bad_message(self):
        f = mock.Mock()
        alsoProvides(f, IFramer)
        f.add_and_parse = mock.Mock(return_value=[Prologue(),
                                                  Frame(frame=b"rx-handshake"),
                                                  Frame(frame=b"bad-message")])
        n = mock.Mock()
        n.write_message = mock.Mock(return_value=b"tx-handshake")
        nvm = NoiseInvalidMessage()
        n.decrypt = mock.Mock(side_effect=nvm)
        r = _Record(f, n, LEADER)
        r.set_role_leader()
        self.assertEqual(f.mock_calls, [])
        r.connectionMade()
        self.assertEqual(f.mock_calls, [mock.call.connectionMade()])
        f.mock_calls[:] = []
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        n.mock_calls[:] = []

        with mock.patch("wormhole._dilation.connection.log.err") as le:
            with self.assertRaises(Disconnect):
                list(r.add_and_unframe(b"data"))
        self.assertEqual(le.mock_calls,
                         [mock.call(nvm, "bad inbound noise frame")])

    def test_send_record(self):
        f = mock.Mock()
        alsoProvides(f, IFramer)
        n = mock.Mock()
        f1 = b"some bytes"
        n.encrypt = mock.Mock(return_value=f1)
        r1 = Ping(b"pingid")
        r = _Record(f, n, LEADER)
        r.set_role_leader()
        self.assertEqual(f.mock_calls, [])
        m1 = b"some other bytes"
        with mock.patch("wormhole._dilation.connection.encode_record",
                        return_value=m1) as er:
            r.send_record(r1)
        self.assertEqual(er.mock_calls, [mock.call(r1)])
        self.assertEqual(n.mock_calls, [mock.call.start_handshake(),
                                        mock.call.encrypt(m1)])
        self.assertEqual(f.mock_calls, [mock.call.send_frame(f1)])

    def test_good(self):
        # Exercise the success path. The Record instance is given each chunk
        # of data as it arrives on Protocol.dataReceived, and is supposed to
        # return a series of Tokens (maybe none, if the chunk was incomplete,
        # or more than one, if the chunk was larger). Internally, it delivers
        # the chunks to the Framer for unframing (which returns 0 or more
        # frames), manages the Noise decryption object, and parses any
        # decrypted messages into tokens (some of which are consumed
        # internally, others for delivery upstairs).
        #
        # in the normal flow, we get:
        #
        # |   | Inbound   | NoiseAction   | Outbound  | ToUpstairs |
        # |   | -         | -             | -         | -          |
        # | 1 |           |               | prologue  |            |
        # | 2 | prologue  |               |           |            |
        # | 3 |           | write_message | handshake |            |
        # | 4 | handshake | read_message  |           | Handshake  |
        # | 5 |           | encrypt       | KCM       |            |
        # | 6 | KCM       | decrypt       |           | KCM        |
        # | 7 | msg1      | decrypt       |           | msg1       |

        # 1: instantiating the Record instance causes the outbound prologue
        # to be sent

        # 2+3: receipt of the inbound prologue triggers creation of the
        # ephemeral key (the "handshake") by calling noise.write_message()
        # and then writes the handshake to the outbound transport

        # 4: when the peer's handshake is received, it is delivered to
        # noise.read_message(), which generates the shared key (enabling
        # noise.send() and noise.decrypt()). It also delivers the Handshake
        # token upstairs, which might (on the Follower) trigger immediate
        # transmission of the Key Confirmation Message (KCM)

        # 5: the outbound KCM is framed and fed into noise.encrypt(), then
        # sent outbound

        # 6: the peer's KCM is decrypted then delivered upstairs. The
        # Follower treats this as a signal that it should use this connection
        # (and drop all others).

        # 7: the peer's first message is decrypted, parsed, and delivered
        # upstairs. This might be an Open or a Data, depending upon what
        # queued messages were left over from the previous connection

        r, f, n = make_record()
        outbound_handshake = object()
        kcm, msg1 = object(), object()
        f_kcm, f_msg1 = object(), object()
        n.write_message = mock.Mock(return_value=outbound_handshake)
        n.decrypt = mock.Mock(side_effect=[kcm, msg1])
        n.encrypt = mock.Mock(side_effect=[f_kcm, f_msg1])
        f.add_and_parse = mock.Mock(side_effect=[[],  # no tokens yet
                                                 [Prologue()],
                                                 [Frame("f_handshake")],
                                                 [Frame("f_kcm"),
                                                  Frame("f_msg1")],
                                                 ])

        self.assertEqual(f.mock_calls, [])
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        n.mock_calls[:] = []

        # 1. The Framer is responsible for sending the prologue, so we don't
        # have to check that here, we just check that the Framer was told
        # about connectionMade properly.
        r.connectionMade()
        self.assertEqual(f.mock_calls, [mock.call.connectionMade()])
        self.assertEqual(n.mock_calls, [])
        f.mock_calls[:] = []

        # 2
        # we dribble the prologue in over two messages, to make sure we can
        # handle a dataReceived that doesn't complete the token

        # remember, add_and_unframe is a generator
        self.assertEqual(list(r.add_and_unframe(b"pro")), [])
        self.assertEqual(f.mock_calls, [mock.call.add_and_parse(b"pro")])
        self.assertEqual(n.mock_calls, [])
        f.mock_calls[:] = []

        self.assertEqual(list(r.add_and_unframe(b"logue")), [])
        # 3: write_message, send outbound handshake
        self.assertEqual(f.mock_calls, [mock.call.add_and_parse(b"logue"),
                                        mock.call.send_frame(outbound_handshake),
                                        ])
        self.assertEqual(n.mock_calls, [mock.call.write_message()])
        f.mock_calls[:] = []
        n.mock_calls[:] = []

        # 4
        # Now deliver the Noise "handshake", the ephemeral public key. This
        # is framed, but not a record, so it shouldn't decrypt or parse
        # anything, but the handshake is delivered to the Noise object, and
        # it does return a Handshake token so we can let the next layer up
        # react (by sending the KCM frame if we're a Follower, or not if
        # we're the Leader)

        self.assertEqual(list(r.add_and_unframe(b"handshake")), [Handshake()])
        self.assertEqual(f.mock_calls, [mock.call.add_and_parse(b"handshake")])
        self.assertEqual(n.mock_calls, [mock.call.read_message("f_handshake")])
        f.mock_calls[:] = []
        n.mock_calls[:] = []

        # 5: at this point we ought to be able to send a message, the KCM
        with mock.patch("wormhole._dilation.connection.encode_record",
                        side_effect=[b"r-kcm"]) as er:
            r.send_record(kcm)
        self.assertEqual(er.mock_calls, [mock.call(kcm)])
        self.assertEqual(n.mock_calls, [mock.call.encrypt(b"r-kcm")])
        self.assertEqual(f.mock_calls, [mock.call.send_frame(f_kcm)])
        n.mock_calls[:] = []
        f.mock_calls[:] = []

        # 6: Now we deliver two messages stacked up: the KCM (Key
        # Confirmation Message) and the first real message. Concatenating
        # them tests that we can handle more than one token in a single
        # chunk. We need to mock parse_record() because everything past the
        # handshake is decrypted and parsed.

        with mock.patch("wormhole._dilation.connection.parse_record",
                        side_effect=[kcm, msg1]) as pr:
            self.assertEqual(list(r.add_and_unframe(b"kcm,msg1")),
                             [kcm, msg1])
            self.assertEqual(f.mock_calls,
                             [mock.call.add_and_parse(b"kcm,msg1")])
            self.assertEqual(n.mock_calls, [mock.call.decrypt("f_kcm"),
                                            mock.call.decrypt("f_msg1")])
            self.assertEqual(pr.mock_calls, [mock.call(kcm), mock.call(msg1)])
        n.mock_calls[:] = []
        f.mock_calls[:] = []

    def test_large_frame(self):
        """
        Noise only allows 64KiB message, but the API allows up to 4GiB
        frames
        """
        # XXX could really benefit from some Hypothesis style
        # exploration of more cases .. but we don't already depend on
        # that library, so a future improvement

        @implementer(ITransport)
        class FakeTransport:
            """
            Record which write()s happen
            """
            def __init__(self):
                self.data = []  # list of bytes

            def write(self, data):
                self.data.append(data)

        # we build both sides of a connection so that underlying Noise
        # structures can be set up and paired properly. Essentially
        # this test is acting like the L2 Protocol object, and can
        # feed bytes to / from either side
        pake_secret = b"\x00" * 32
        transport0 = FakeTransport()
        transport1 = FakeTransport()
        noise0 = build_noise()
        noise0.set_psks(pake_secret)
        noise0.set_as_initiator()  # leader
        noise1 = build_noise()
        noise1.set_psks(pake_secret)
        noise1.set_as_responder()  # follower
        framer0 = _Framer(transport0, b"out prolog", b"in prolog")
        record0 = _Record(framer0, noise0, LEADER)
        record0.set_role_leader()
        record0.connectionMade()
        self.assertEqual(
            list(record0.add_and_unframe(b"in prolog")),
            []
        )

        # note that in connector the prologues flip around depending
        # on who is the leader or follower
        framer1 = _Framer(transport1, b"in prolog", b"out prolog")
        record1 = _Record(framer1, noise1, FOLLOWER)
        record1.set_role_follower()
        record1.connectionMade()

        # the leader has now sent the prolog and the opening handshake
        # message, so we consume them on the follower side

        self.assertEqual(
            list(record1.add_and_unframe(transport0.data[0])),  # b"out prolog"
            []
        )
        self.assertEqual(
            list(record1.add_and_unframe(transport0.data[1])),
            [Handshake()]
        )

        # the follower sends the first KCM; the leader is waiting for
        # this and will complete the handshake after that
        record1.send_record(KCM())
        self.assertEqual(
            list(record0.add_and_unframe(transport1.data[1])),
            [Handshake()]
        )
        record0.send_record(KCM())
        self.assertEqual(
            list(record1.add_and_unframe(transport0.data[2])),
            [KCM()]
        )
        # both sides are now done their handshakes

        # Now, we send a message that's definitely bigger than a
        # single Noise message can deal with
        input_plaintext = b"\xff" * 65537
        record0.send_record(
            Data(
                seqnum=123,
                scid=456,
                data=input_plaintext,
            )
        )

        # ...despite its size the message should still be sent out as
        # a single Data, because _our_ framing handles 4-byte lengths
        self.assertEqual(
            len(transport0.data[3]),
            4 + 65537 + (16 * 2) + 1 + 4 + 4
        )
        #   ^-- frame-length
        #       ^-- plaintext size
        #               ^-- Noise associated data, x2 noise packets
        #                          ^-- "data" kind, 0x04
        #                              ^-- subchannel-id
        #                                  ^-- sequence number
        #
        # (maybe we don't need the above assert, since if any of that
        # isn't true the next step will fail)

        # round-trip the data to the other side and ensure we get the
        # plaintext back
        outputs = list(
            record1.add_and_unframe(transport0.data[3])
        )
        self.assertEqual(len(outputs), 1)
        self.assertEqual(
            outputs[0],
            Data(
                seqnum=123,
                scid=456,
                data=input_plaintext,
            )
        )



import re
from unittest import mock
from twisted.internet import reactor
from twisted.trial import unittest
from twisted.internet.task import Cooperator
from twisted.internet.defer import Deferred, inlineCallbacks
from zope.interface import implementer

from ... import _interfaces
from ...eventual import EventualQueue
from ..._interfaces import ITerminator
from ..._dilation import manager
from ..._dilation._noise import NoiseConnection


@implementer(_interfaces.ISend)
class MySend(object):
    def __init__(self, side):
        self.rx_phase = 0
        self.side = side

    def send(self, phase, plaintext):
        # print("SEND[%s]" % self.side, phase, plaintext)
        self.peer.got(phase, plaintext)

    def got(self, phase, plaintext):
        d_mo = re.search(r'^dilate-(\d+)$', phase)
        p = int(d_mo.group(1))
        assert p == self.rx_phase
        self.rx_phase += 1
        self.dilator.received_dilate(plaintext)


@implementer(ITerminator)
class FakeTerminator(object):
    def __init__(self):
        self.d = Deferred()

    def stoppedD(self):
        self.d.callback(None)


class Connect(unittest.TestCase):
    @inlineCallbacks
    def test1(self):
        if not NoiseConnection:
            raise unittest.SkipTest("noiseprotocol unavailable")
        # print()
        send_left = MySend("left")
        send_right = MySend("right")
        send_left.peer = send_right
        send_right.peer = send_left
        key = b"\x00"*32
        eq = EventualQueue(reactor)
        cooperator = Cooperator(scheduler=eq.eventually)

        t_left = FakeTerminator()
        t_right = FakeTerminator()

        d_left = manager.Dilator(reactor, eq, cooperator)
        d_left.wire(send_left, t_left)
        d_left.got_key(key)
        d_left.got_wormhole_versions({"can-dilate": ["1"]})
        send_left.dilator = d_left

        d_right = manager.Dilator(reactor, eq, cooperator)
        d_right.wire(send_right, t_right)
        d_right.got_key(key)
        d_right.got_wormhole_versions({"can-dilate": ["1"]})
        send_right.dilator = d_right

        with mock.patch("wormhole._dilation.connector.ipaddrs.find_addresses",
                        return_value=["127.0.0.1"]):
            eps_left_d = d_left.dilate(no_listen=True)
            eps_right_d = d_right.dilate()

        eps_left = yield eps_left_d
        eps_right = yield eps_right_d

        # print("left connected", eps_left)
        # print("right connected", eps_right)

        control_ep_left, connect_ep_left, listen_ep_left = eps_left
        control_ep_right, connect_ep_right, listen_ep_right = eps_right

        # we normally shut down with w.close(), which calls Dilator.stop(),
        # which calls Terminator.stoppedD(), which (after everything else is
        # done) calls Boss.stopped
        d_left.stop()
        d_right.stop()

        yield t_left.d
        yield t_right.d



from twisted.internet.task import Clock
from twisted.python.failure import Failure
from twisted.trial import unittest

from ..eventual import EventualQueue
from ..observer import OneShotObserver, SequenceObserver, EmptyableSet


class OneShot(unittest.TestCase):
    def test_fire(self):
        c = Clock()
        eq = EventualQueue(c)
        o = OneShotObserver(eq)
        res = object()
        d1 = o.when_fired()
        eq.flush_sync()
        self.assertNoResult(d1)
        o.fire(res)
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d1), res)
        d2 = o.when_fired()
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d2), res)
        o.fire_if_not_fired(object())
        eq.flush_sync()

    def test_fire_if_not_fired(self):
        c = Clock()
        eq = EventualQueue(c)
        o = OneShotObserver(eq)
        res1 = object()
        res2 = object()
        d1 = o.when_fired()
        eq.flush_sync()
        self.assertNoResult(d1)
        o.fire_if_not_fired(res1)
        o.fire_if_not_fired(res2)
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d1), res1)

    def test_error_before_firing(self):
        c = Clock()
        eq = EventualQueue(c)
        o = OneShotObserver(eq)
        f = Failure(ValueError("oops"))
        d1 = o.when_fired()
        eq.flush_sync()
        self.assertNoResult(d1)
        o.error(f)
        eq.flush_sync()
        self.assertIdentical(self.failureResultOf(d1), f)
        d2 = o.when_fired()
        eq.flush_sync()
        self.assertIdentical(self.failureResultOf(d2), f)

    def test_error_after_firing(self):
        c = Clock()
        eq = EventualQueue(c)
        o = OneShotObserver(eq)
        res = object()
        f = Failure(ValueError("oops"))

        o.fire(res)
        eq.flush_sync()
        d1 = o.when_fired()
        o.error(f)
        d2 = o.when_fired()
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d1), res)
        self.assertIdentical(self.failureResultOf(d2), f)


class Sequence(unittest.TestCase):
    def test_fire(self):
        c = Clock()
        eq = EventualQueue(c)
        o = SequenceObserver(eq)
        d1 = o.when_next_event()
        eq.flush_sync()
        self.assertNoResult(d1)
        d2 = o.when_next_event()
        eq.flush_sync()
        self.assertNoResult(d1)
        self.assertNoResult(d2)

        ev1 = object()
        ev2 = object()
        o.fire(ev1)
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d1), ev1)
        self.assertNoResult(d2)

        o.fire(ev2)
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d2), ev2)

        ev3 = object()
        ev4 = object()
        o.fire(ev3)
        o.fire(ev4)

        d3 = o.when_next_event()
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d3), ev3)

        d4 = o.when_next_event()
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d4), ev4)

    def test_error(self):
        c = Clock()
        eq = EventualQueue(c)
        o = SequenceObserver(eq)
        d1 = o.when_next_event()
        eq.flush_sync()
        self.assertNoResult(d1)
        f = Failure(ValueError("oops"))
        o.fire(f)
        eq.flush_sync()
        self.assertIdentical(self.failureResultOf(d1), f)
        d2 = o.when_next_event()
        eq.flush_sync()
        self.assertIdentical(self.failureResultOf(d2), f)


class Empty(unittest.TestCase):
    def test_set(self):
        eq = EventualQueue(Clock())
        s = EmptyableSet(_eventual_queue=eq)
        d1 = s.when_next_empty()
        eq.flush_sync()
        self.assertNoResult(d1)
        s.add(1)
        eq.flush_sync()
        self.assertNoResult(d1)
        s.add(2)
        s.discard(1)
        d2 = s.when_next_empty()
        eq.flush_sync()
        self.assertNoResult(d1)
        self.assertNoResult(d2)
        s.discard(2)
        eq.flush_sync()
        self.assertEqual(self.successResultOf(d1), None)
        self.assertEqual(self.successResultOf(d2), None)

        s.add(3)
        s.discard(3)



from __future__ import print_function, unicode_literals
from twisted.trial import unittest
from ..._dilation.encode import to_be4, from_be4


class Encoding(unittest.TestCase):

    def test_be4(self):
        self.assertEqual(to_be4(0), b"\x00\x00\x00\x00")
        self.assertEqual(to_be4(1), b"\x00\x00\x00\x01")
        self.assertEqual(to_be4(256), b"\x00\x00\x01\x00")
        self.assertEqual(to_be4(257), b"\x00\x00\x01\x01")
        with self.assertRaises(ValueError):
            to_be4(-1)
        with self.assertRaises(ValueError):
            to_be4(2**32)

        self.assertEqual(from_be4(b"\x00\x00\x00\x00"), 0)
        self.assertEqual(from_be4(b"\x00\x00\x00\x01"), 1)
        self.assertEqual(from_be4(b"\x00\x00\x01\x00"), 256)
        self.assertEqual(from_be4(b"\x00\x00\x01\x01"), 257)

        with self.assertRaises(TypeError):
            from_be4(0)
        with self.assertRaises(ValueError):
            from_be4(b"\x01\x00\x00\x00\x00")



from __future__ import print_function, unicode_literals

from unittest import mock
from zope.interface import alsoProvides
from twisted.trial import unittest
from twisted.internet.task import Clock
from twisted.internet.defer import Deferred
from twisted.internet.address import IPv4Address, IPv6Address, HostnameAddress
from ...eventual import EventualQueue
from ..._interfaces import IDilationManager, IDilationConnector
from ..._hints import DirectTCPV1Hint, RelayV1Hint, TorTCPV1Hint
from ..._dilation import roles
from ..._dilation._noise import NoiseConnection
from ..._dilation.connection import KCM
from ..._dilation.connector import (Connector,
                                    build_sided_relay_handshake,
                                    build_noise,
                                    describe_inbound,
                                    OutboundConnectionFactory,
                                    InboundConnectionFactory,
                                    PROLOGUE_LEADER, PROLOGUE_FOLLOWER,
                                    )
from .common import clear_mock_calls


class Handshake(unittest.TestCase):
    def test_build(self):
        key = b"k"*32
        side = "12345678abcdabcd"
        self.assertEqual(
            build_sided_relay_handshake(key, side),
            b"please relay 3f4147851dbd2589d25b654ee9fb35ed0d3e5f19c5c5403e8e6a195c70f0577a"
            b" for side 12345678abcdabcd\n"
        )


class Outbound(unittest.TestCase):
    def test_no_relay(self):
        c = mock.Mock()
        alsoProvides(c, IDilationConnector)
        p0 = mock.Mock()
        c.build_protocol = mock.Mock(return_value=p0)
        relay_handshake = None
        f = OutboundConnectionFactory(c, relay_handshake, "desc")
        addr = object()
        p = f.buildProtocol(addr)
        self.assertIdentical(p, p0)
        self.assertEqual(c.mock_calls, [mock.call.build_protocol(addr, "desc")])
        self.assertEqual(p.mock_calls, [])
        self.assertIdentical(p.factory, f)

    def test_with_relay(self):
        c = mock.Mock()
        alsoProvides(c, IDilationConnector)
        p0 = mock.Mock()
        c.build_protocol = mock.Mock(return_value=p0)
        relay_handshake = b"relay handshake"
        f = OutboundConnectionFactory(c, relay_handshake, "desc")
        addr = object()
        p = f.buildProtocol(addr)
        self.assertIdentical(p, p0)
        self.assertEqual(c.mock_calls, [mock.call.build_protocol(addr, "desc")])
        self.assertEqual(p.mock_calls, [mock.call.use_relay(relay_handshake)])
        self.assertIdentical(p.factory, f)


class Inbound(unittest.TestCase):
    def test_build(self):
        c = mock.Mock()
        alsoProvides(c, IDilationConnector)
        p0 = mock.Mock()
        c.build_protocol = mock.Mock(return_value=p0)
        f = InboundConnectionFactory(c)
        addr = IPv4Address("TCP", "1.2.3.4", 55)
        p = f.buildProtocol(addr)
        self.assertIdentical(p, p0)
        self.assertEqual(c.mock_calls, [mock.call.build_protocol(addr, "<-tcp:1.2.3.4:55")])
        self.assertIdentical(p.factory, f)


def make_connector(listen=True, tor=False, relay=None, role=roles.LEADER):
    class Holder:
        pass
    h = Holder()
    h.dilation_key = b"key"
    h.relay = relay
    h.manager = mock.Mock()
    alsoProvides(h.manager, IDilationManager)
    h.clock = Clock()
    h.reactor = h.clock
    h.eq = EventualQueue(h.clock)
    h.tor = None
    if tor:
        h.tor = mock.Mock()
    timing = None
    h.side = u"abcd1234abcd5678"
    h.role = role
    c = Connector(h.dilation_key, h.relay, h.manager, h.reactor, h.eq,
                  not listen, h.tor, timing, h.side, h.role)
    return c, h


class TestConnector(unittest.TestCase):
    def test_build(self):
        c, h = make_connector()
        c, h = make_connector(relay="tcp:host:1234")

    def test_connection_abilities(self):
        self.assertEqual(Connector.get_connection_abilities(),
                         [{"type": "direct-tcp-v1"},
                          {"type": "relay-v1"},
                          ])

    def test_build_noise(self):
        if not NoiseConnection:
            raise unittest.SkipTest("noiseprotocol unavailable")
        build_noise()

    def test_build_protocol_leader(self):
        c, h = make_connector(role=roles.LEADER)
        n0 = mock.Mock()
        p0 = mock.Mock()
        addr = object()
        with mock.patch("wormhole._dilation.connector.build_noise",
                        return_value=n0) as bn:
            with mock.patch("wormhole._dilation.connector.DilatedConnectionProtocol",
                            return_value=p0) as dcp:
                p = c.build_protocol(addr, "desc")
        self.assertEqual(bn.mock_calls, [mock.call()])
        self.assertEqual(n0.mock_calls, [mock.call.set_psks(h.dilation_key),
                                         mock.call.set_as_initiator()])
        self.assertIdentical(p, p0)
        self.assertEqual(dcp.mock_calls,
                         [mock.call(h.eq, h.role, "desc", c, n0,
                                    PROLOGUE_LEADER, PROLOGUE_FOLLOWER)])

    def test_build_protocol_follower(self):
        c, h = make_connector(role=roles.FOLLOWER)
        n0 = mock.Mock()
        p0 = mock.Mock()
        addr = object()
        with mock.patch("wormhole._dilation.connector.build_noise",
                        return_value=n0) as bn:
            with mock.patch("wormhole._dilation.connector.DilatedConnectionProtocol",
                            return_value=p0) as dcp:
                p = c.build_protocol(addr, "desc")
        self.assertEqual(bn.mock_calls, [mock.call()])
        self.assertEqual(n0.mock_calls, [mock.call.set_psks(h.dilation_key),
                                         mock.call.set_as_responder()])
        self.assertIdentical(p, p0)
        self.assertEqual(dcp.mock_calls,
                         [mock.call(h.eq, h.role, "desc", c, n0,
                                    PROLOGUE_FOLLOWER, PROLOGUE_LEADER)])

    def test_start_stop(self):
        c, h = make_connector(listen=False, relay=None, role=roles.LEADER)
        c.start()
        # no relays, so it publishes no hints
        self.assertEqual(h.manager.mock_calls, [])
        # and no listener, so nothing happens until we provide a hint
        c.stop()
        # we stop while we're connecting, so no connections must be stopped

    def test_empty(self):
        c, h = make_connector(listen=False, relay=None, role=roles.LEADER)
        c._schedule_connection = mock.Mock()
        c.start()
        # no relays, so it publishes no hints
        self.assertEqual(h.manager.mock_calls, [])
        # and no listener, so nothing happens until we provide a hint
        self.assertEqual(c._schedule_connection.mock_calls, [])
        c.stop()

    def test_basic(self):
        c, h = make_connector(listen=False, relay=None, role=roles.LEADER)
        c._schedule_connection = mock.Mock()
        c.start()
        # no relays, so it publishes no hints
        self.assertEqual(h.manager.mock_calls, [])
        # and no listener, so nothing happens until we provide a hint
        self.assertEqual(c._schedule_connection.mock_calls, [])

        hint = DirectTCPV1Hint("foo", 55, 0.0)
        c.got_hints([hint])

        # received hints don't get published
        self.assertEqual(h.manager.mock_calls, [])
        # they just schedule a connection
        self.assertEqual(c._schedule_connection.mock_calls,
                         [mock.call(0.0, DirectTCPV1Hint("foo", 55, 0.0),
                                    is_relay=False)])

    def test_listen_addresses(self):
        c, h = make_connector(listen=True, role=roles.LEADER)
        with mock.patch("wormhole.ipaddrs.find_addresses",
                        return_value=["127.0.0.1", "1.2.3.4", "5.6.7.8"]):
            self.assertEqual(c._get_listener_addresses(),
                             ["1.2.3.4", "5.6.7.8"])
        with mock.patch("wormhole.ipaddrs.find_addresses",
                        return_value=["127.0.0.1"]):
            # some test hosts, including the appveyor VMs, *only* have
            # 127.0.0.1, and the tests will hang badly if we remove it.
            self.assertEqual(c._get_listener_addresses(), ["127.0.0.1"])

    def test_listen(self):
        c, h = make_connector(listen=True, role=roles.LEADER)
        c._start_listener = mock.Mock()
        with mock.patch("wormhole.ipaddrs.find_addresses",
                        return_value=["127.0.0.1", "1.2.3.4", "5.6.7.8"]):
            c.start()
        self.assertEqual(c._start_listener.mock_calls,
                         [mock.call(["1.2.3.4", "5.6.7.8"])])

    def test_start_listen(self):
        c, h = make_connector(listen=True, role=roles.LEADER)
        ep = mock.Mock()
        d = Deferred()
        ep.listen = mock.Mock(return_value=d)
        with mock.patch("wormhole._dilation.connector.serverFromString",
                        return_value=ep) as sfs:
            c._start_listener(["1.2.3.4", "5.6.7.8"])
        self.assertEqual(sfs.mock_calls, [mock.call(h.reactor, "tcp:0")])
        lp = mock.Mock()
        host = mock.Mock()
        host.port = 66
        lp.getHost = mock.Mock(return_value=host)
        d.callback(lp)
        self.assertEqual(h.manager.mock_calls,
                         [mock.call.send_hints([{"type": "direct-tcp-v1",
                                                 "hostname": "1.2.3.4",
                                                 "port": 66,
                                                 "priority": 0.0
                                                 },
                                                {"type": "direct-tcp-v1",
                                                 "hostname": "5.6.7.8",
                                                 "port": 66,
                                                 "priority": 0.0
                                                 },
                                                ])])

    def test_schedule_connection_no_relay(self):
        c, h = make_connector(listen=True, role=roles.LEADER)
        hint = DirectTCPV1Hint("foo", 55, 0.0)
        ep = mock.Mock()
        with mock.patch("wormhole._dilation.connector.endpoint_from_hint_obj",
                        side_effect=[ep]) as efho:
            c._schedule_connection(0.0, hint, False)
        self.assertEqual(efho.mock_calls, [mock.call(hint, h.tor, h.reactor)])
        self.assertEqual(ep.mock_calls, [])
        d = Deferred()
        ep.connect = mock.Mock(side_effect=[d])
        # direct hints are scheduled for T+0.0
        f = mock.Mock()
        with mock.patch("wormhole._dilation.connector.OutboundConnectionFactory",
                        return_value=f) as ocf:
            h.clock.advance(1.0)
        self.assertEqual(ocf.mock_calls, [mock.call(c, None, "->tcp:foo:55")])
        self.assertEqual(ep.connect.mock_calls, [mock.call(f)])
        p = mock.Mock()
        d.callback(p)
        self.assertEqual(p.mock_calls,
                         [mock.call.when_disconnected(),
                          mock.call.when_disconnected().addCallback(c._pending_connections.discard)])

    def test_schedule_connection_relay(self):
        c, h = make_connector(listen=True, role=roles.LEADER)
        hint = DirectTCPV1Hint("foo", 55, 0.0)
        ep = mock.Mock()
        with mock.patch("wormhole._dilation.connector.endpoint_from_hint_obj",
                        side_effect=[ep]) as efho:
            c._schedule_connection(0.0, hint, True)
        self.assertEqual(efho.mock_calls, [mock.call(hint, h.tor, h.reactor)])
        self.assertEqual(ep.mock_calls, [])
        d = Deferred()
        ep.connect = mock.Mock(side_effect=[d])
        # direct hints are scheduled for T+0.0
        f = mock.Mock()
        with mock.patch("wormhole._dilation.connector.OutboundConnectionFactory",
                        return_value=f) as ocf:
            h.clock.advance(1.0)
        handshake = build_sided_relay_handshake(h.dilation_key, h.side)
        self.assertEqual(ocf.mock_calls, [mock.call(c, handshake, "->relay:tcp:foo:55")])

    def test_listen_but_tor(self):
        c, h = make_connector(listen=True, tor=True, role=roles.LEADER)
        with mock.patch("wormhole.ipaddrs.find_addresses",
                        return_value=["127.0.0.1", "1.2.3.4", "5.6.7.8"]) as fa:
            c.start()
        # don't even look up addresses
        self.assertEqual(fa.mock_calls, [])
        # no relays and the listener isn't ready yet, so no hints yet
        self.assertEqual(h.manager.mock_calls, [])

    def test_no_listen(self):
        c, h = make_connector(listen=False, tor=False, role=roles.LEADER)
        with mock.patch("wormhole.ipaddrs.find_addresses",
                        return_value=["127.0.0.1", "1.2.3.4", "5.6.7.8"]) as fa:
            c.start()
        # don't even look up addresses
        self.assertEqual(fa.mock_calls, [])
        self.assertEqual(h.manager.mock_calls, [])

    def test_relay_delay(self):
        # given a direct connection and a relay, we should see the direct
        # connection initiated at T+0 seconds, and the relay at T+RELAY_DELAY
        c, h = make_connector(listen=True, relay=None, role=roles.LEADER)
        c._schedule_connection = mock.Mock()
        c._start_listener = mock.Mock()
        c.start()
        hint1 = DirectTCPV1Hint("foo", 55, 0.0)
        hint2 = DirectTCPV1Hint("bar", 55, 0.0)
        hint3 = RelayV1Hint([DirectTCPV1Hint("relay", 55, 0.0)])
        c.got_hints([hint1, hint2, hint3])
        self.assertEqual(c._schedule_connection.mock_calls,
                         [mock.call(0.0, hint1, is_relay=False),
                          mock.call(0.0, hint2, is_relay=False),
                          mock.call(c.RELAY_DELAY, hint3.hints[0], is_relay=True),
                          ])

    def test_initial_relay(self):
        c, h = make_connector(listen=False, relay="tcp:foo:55", role=roles.LEADER)
        c._schedule_connection = mock.Mock()
        c.start()
        self.assertEqual(h.manager.mock_calls,
                         [mock.call.send_hints([{"type": "relay-v1",
                                                 "hints": [
                                                     {"type": "direct-tcp-v1",
                                                      "hostname": "foo",
                                                      "port": 55,
                                                      "priority": 0.0
                                                      },
                                                     ],
                                                 }])])
        self.assertEqual(c._schedule_connection.mock_calls,
                         [mock.call(0.0, DirectTCPV1Hint("foo", 55, 0.0),
                                    is_relay=True)])

    def test_tor_no_manager(self):
        # tor hints should be ignored if we don't have a Tor manager to use them
        c, h = make_connector(listen=False, role=roles.LEADER)
        c._schedule_connection = mock.Mock()
        c.start()
        hint = TorTCPV1Hint("foo", 55, 0.0)
        c.got_hints([hint])
        self.assertEqual(h.manager.mock_calls, [])
        self.assertEqual(c._schedule_connection.mock_calls, [])

    def test_tor_with_manager(self):
        # tor hints should be processed if we do have a Tor manager
        c, h = make_connector(listen=False, tor=True, role=roles.LEADER)
        c._schedule_connection = mock.Mock()
        c.start()
        hint = TorTCPV1Hint("foo", 55, 0.0)
        c.got_hints([hint])
        self.assertEqual(c._schedule_connection.mock_calls,
                         [mock.call(0.0, hint, is_relay=False)])

    def test_priorities(self):
        # given two hints with different priorities, we should somehow prefer
        # one. This is a placeholder to fill in once we implement priorities.
        pass


class Race(unittest.TestCase):
    def test_one_leader(self):
        c, h = make_connector(listen=True, role=roles.LEADER)
        lp = mock.Mock()

        def start_listener(addresses):
            c._listeners.add(lp)
        c._start_listener = start_listener
        c._schedule_connection = mock.Mock()
        c.start()
        self.assertEqual(c._listeners, set([lp]))

        p1 = mock.Mock()  # DilatedConnectionProtocol instance
        c.add_candidate(p1)
        self.assertEqual(h.manager.mock_calls, [])
        h.eq.flush_sync()
        self.assertEqual(h.manager.mock_calls, [mock.call.connector_connection_made(p1)])
        self.assertEqual(p1.mock_calls,
                         [mock.call.select(h.manager),
                          mock.call.send_record(KCM())])
        self.assertEqual(lp.mock_calls[0], mock.call.stopListening())
        # stop_listeners() uses a DeferredList, so we ignore the second call

    def test_one_follower(self):
        c, h = make_connector(listen=True, role=roles.FOLLOWER)
        lp = mock.Mock()

        def start_listener(addresses):
            c._listeners.add(lp)
        c._start_listener = start_listener
        c._schedule_connection = mock.Mock()
        c.start()
        self.assertEqual(c._listeners, set([lp]))

        p1 = mock.Mock()  # DilatedConnectionProtocol instance
        c.add_candidate(p1)
        self.assertEqual(h.manager.mock_calls, [])
        h.eq.flush_sync()
        self.assertEqual(h.manager.mock_calls, [mock.call.connector_connection_made(p1)])
        # just like LEADER, but follower doesn't send KCM now (it sent one
        # earlier, to tell the leader that this connection looks viable)
        self.assertEqual(p1.mock_calls,
                         [mock.call.select(h.manager)])
        self.assertEqual(lp.mock_calls[0], mock.call.stopListening())
        # stop_listeners() uses a DeferredList, so we ignore the second call

    # TODO: make sure a pending connection is abandoned when the listener
    # answers successfully

    # TODO: make sure a second pending connection is abandoned when the first
    # connection succeeds

    def test_late(self):
        c, h = make_connector(listen=False, role=roles.LEADER)
        c._schedule_connection = mock.Mock()
        c.start()

        p1 = mock.Mock()  # DilatedConnectionProtocol instance
        c.add_candidate(p1)
        self.assertEqual(h.manager.mock_calls, [])
        h.eq.flush_sync()
        self.assertEqual(h.manager.mock_calls, [mock.call.connector_connection_made(p1)])
        clear_mock_calls(h.manager)
        self.assertEqual(p1.mock_calls,
                         [mock.call.select(h.manager),
                          mock.call.send_record(KCM())])

        # late connection is ignored
        p2 = mock.Mock()
        c.add_candidate(p2)
        self.assertEqual(h.manager.mock_calls, [])

    # make sure an established connection is dropped when stop() is called
    def test_stop(self):
        c, h = make_connector(listen=False, role=roles.LEADER)
        c._schedule_connection = mock.Mock()
        c.start()

        p1 = mock.Mock()  # DilatedConnectionProtocol instance
        c.add_candidate(p1)
        self.assertEqual(h.manager.mock_calls, [])
        h.eq.flush_sync()
        self.assertEqual(p1.mock_calls,
                         [mock.call.select(h.manager),
                          mock.call.send_record(KCM())])
        self.assertEqual(h.manager.mock_calls, [mock.call.connector_connection_made(p1)])

        c.stop()


class Describe(unittest.TestCase):
    def test_describe_inbound(self):
        self.assertEqual(describe_inbound(HostnameAddress("example.com", 1234)),
                         "<-tcp:example.com:1234")
        self.assertEqual(describe_inbound(IPv4Address("TCP", "1.2.3.4", 1234)),
                         "<-tcp:1.2.3.4:1234")
        self.assertEqual(describe_inbound(IPv6Address("TCP", "::1", 1234)),
                         "<-tcp:[::1]:1234")
        other = "none-of-the-above"
        self.assertEqual(describe_inbound(other), "<-%r" % other)



from __future__ import print_function, unicode_literals
from collections import deque
from attr import attrs, attrib
from attr.validators import provides
from zope.interface import implementer
from twisted.internet.interfaces import IPushProducer, IPullProducer
from twisted.python import log
from twisted.python.reflect import safe_str
from .._interfaces import IDilationManager, IOutbound
from .connection import KCM, Ping, Pong, Ack


# Outbound flow control: app writes to subchannel, we write to Connection

# The app can register an IProducer of their choice, to let us throttle their
# outbound data. Not all subchannels will have producers registered, and the
# producer probably won't be the IProtocol instance (it'll be something else
# which feeds data out through the protocol, like a t.p.basic.FileSender). If
# a producerless subchannel writes too much, we won't be able to stop them,
# and we'll keep writing records into the Connection even though it's asked
# us to pause. Likewise, when the connection is down (and we're busily trying
# to reestablish a new one), registered subchannels will be paused, but
# unregistered ones will just dump everything in _outbound_queue, and we'll
# consume memory without bound until they stop.

# We need several things:
#
# * Add each registered IProducer to a list, whose order remains stable. We
#   want fairness under outbound throttling: each time the outbound
#   connection opens up (our resumeProducing method is called), we should let
#   just one producer have an opportunity to do transport.write, and then we
#   should pause them again, and not come back to them until everyone else
#   has gotten a turn. So we want an ordered list of producers to track this
#   rotation.
#
# * Remove the IProducer if/when the protocol uses unregisterProducer
#
# * Remove any registered IProducer when the associated Subchannel is closed.
#   This isn't a problem for normal transports, because usually there's a
#   one-to-one mapping from Protocol to Transport, so when the Transport you
#   forget the only reference to the Producer anyways. Our situation is
#   unusual because we have multiple Subchannels that get merged into the
#   same underlying Connection: each Subchannel's Protocol can register a
#   producer on the Subchannel (which is an ITransport), but that adds it to
#   a set of Producers for the Connection (which is also an ITransport). So
#   if the Subchannel is closed, we need to remove its Producer (if any) even
#   though the Connection remains open.
#
# * Register ourselves as an IPushProducer with each successive Connection
#   object. These connections will come and go, but there will never be more
#   than one. When the connection goes away, pause all our producers. When a
#   new one is established, write all our queued messages, then unpause our
#   producers as we would in resumeProducing.
#
# * Inside our resumeProducing call, we'll cycle through all producers,
#   calling their individual resumeProducing methods one at a time. If they
#   write so much data that the Connection pauses us again, we'll find out
#   because our pauseProducing will be called inside that loop. When that
#   happens, we need to stop looping. If we make it through the whole loop
#   without being paused, then all subchannel Producers are left unpaused,
#   and are free to write whenever they want. During this loop, some
#   Producers will be paused, and others will be resumed
#
# * If our pauseProducing is called, all Producers must be paused, and a flag
#   should be set to notify the resumeProducing loop to exit
#
# * In between calls to our resumeProducing method, we're in one of two
#   states.
#   * If we're writing data too fast, then we'll be left in the "paused"
#     state, in which all Subchannel producers are paused, and the aggregate
#     is paused too (our Connection told us to pauseProducing and hasn't yet
#     told us to resumeProducing). In this state, activity is driven by the
#     outbound TCP window opening up, which calls resumeProducing and allows
#     (probably just) one message to be sent. We receive pauseProducing in
#     the middle of their transport.write, so the loop exits early, and the
#     only state change is that some other Producer should get to go next
#     time.
#   * If we're writing too slowly, we'll be left in the "unpaused" state: all
#     Subchannel producers are unpaused, and the aggregate is unpaused too
#     (resumeProducing is the last thing we've been told). In this state,
#     activity is driven by the Subchannels doing a transport.write, which
#     queues some data on the TCP connection (and then might call
#     pauseProducing if it's now full).
#
# * We want to guard against:
#
#   * application protocol registering a Producer without first unregistering
#     the previous one
#
#   * application protocols writing data despite being told to pause
#     (Subchannels without a registered Producer cannot be throttled, and we
#     can't do anything about that, but we must also handle the case where
#     they give us a pause switch and then proceed to ignore it)
#
#   * our Connection calling resumeProducing or pauseProducing without an
#     intervening call of the other kind
#
#   * application protocols that don't handle a resumeProducing or
#     pauseProducing call without an intervening call of the other kind (i.e.
#     we should keep track of the last thing we told them, and not repeat
#     ourselves)
#
# * If the Wormhole is closed, all Subchannels should close. This is not our
#   responsibility: it lives in (Manager? Inbound?)
#
# * If we're given an IPullProducer, we should keep calling its
#   resumeProducing until it runs out of data. We still want fairness, so we
#   won't call it a second time until everyone else has had a turn.


# There are a couple of different ways to approach this. The one I've
# selected is:
#
# * keep a dict that maps from Subchannel to Producer, which only contains
#   entries for Subchannels that have registered a producer. We use this to
#   remove Producers when Subchannels are closed
#
# * keep a Deque of Producers. This represents the fair-throttling rotation:
#   the left-most item gets the next upcoming turn, and then they'll be moved
#   to the end of the queue.
#
# * keep a set of IPushProducers which are paused, a second set of
#   IPushProducers which are unpaused, and a third set of IPullProducers
#   (which are always left paused) Enforce the invariant that these three
#   sets are disjoint, and that their union equals the contents of the deque.
#
# * keep a "paused" flag, which is cleared upon entry to resumeProducing, and
#   set upon entry to pauseProducing. The loop inside resumeProducing checks
#   this flag after each call to producer.resumeProducing, to sense whether
#   they used their turn to write data, and if that write was large enough to
#   fill the TCP window. If set, we break out of the loop. If not, we look
#   for the next producer to unpause. The loop finishes when all producers
#   are unpaused (evidenced by the two sets of paused producers being empty)
#
# * the "paused" flag also determines whether new IPushProducers are added to
#   the paused or unpaused set (IPullProducers are always added to the
#   pull+paused set). If we have any IPullProducers, we're always in the
#   "writing data too fast" state.

# other approaches that I didn't decide to do at this time (but might use in
# the future):
#
# * use one set instead of two. pros: fewer moving parts. cons: harder to
#   spot decoherence bugs like adding a producer to the deque but forgetting
#   to add it to one of the
#
# * use zero sets, and keep the paused-vs-unpaused state in the Subchannel as
#   a visible boolean flag. This conflates Subchannels with their associated
#   Producer (so if we went this way, we should also let them track their own
#   Producer). Our resumeProducing loop ends when 'not any(sc.paused for sc
#   in self._subchannels_with_producers)'. Pros: fewer subchannel->producer
#   mappings lying around to disagree with one another. Cons: exposes a bit
#   too much of the Subchannel internals


@attrs
@implementer(IOutbound, IPushProducer)
class Outbound(object):
    # Manage outbound data: subchannel writes to us, we write to transport
    _manager = attrib(validator=provides(IDilationManager))
    _cooperator = attrib()

    def __attrs_post_init__(self):
        # _outbound_queue holds all messages we've ever sent but not retired
        self._outbound_queue = deque()
        self._next_outbound_seqnum = 0
        # _queued_unsent are messages to retry with our new connection
        self._queued_unsent = deque()

        # outbound flow control: the Connection throttles our writes
        self._subchannel_producers = {}  # Subchannel -> IProducer
        self._paused = True  # our Connection called our pauseProducing
        self._all_producers = deque()  # rotates, left-is-next
        self._paused_producers = set()
        self._unpaused_producers = set()
        self._check_invariants()

        self._connection = None

    def _check_invariants(self):
        assert self._unpaused_producers.isdisjoint(self._paused_producers)
        assert (self._paused_producers.union(self._unpaused_producers) ==
                set(self._all_producers))

    def build_record(self, record_type, *args):
        seqnum = self._next_outbound_seqnum
        self._next_outbound_seqnum += 1
        r = record_type(seqnum, *args)
        assert hasattr(r, "seqnum"), r  # only Open/Data/Close
        return r

    def queue_and_send_record(self, r):
        # we always queue it, to resend on a subsequent connection if
        # necessary
        self._outbound_queue.append(r)

        if self._connection:
            if self._queued_unsent:
                # to maintain correct ordering, queue this instead of sending it
                self._queued_unsent.append(r)
            else:
                # we're allowed to send it immediately
                self._connection.send_record(r)

    def send_if_connected(self, r):
        assert isinstance(r, (KCM, Ping, Pong, Ack)), r  # nothing with seqnum
        if self._connection:
            self._connection.send_record(r)

    # our subchannels call these to register a producer

    def subchannel_registerProducer(self, sc, producer, streaming):
        # streaming==True: IPushProducer (pause/resume)
        # streaming==False: IPullProducer (just resume)
        if sc in self._subchannel_producers:
            raise ValueError(
                "registering producer %s before previous one (%s) was "
                "unregistered" % (producer,
                                  self._subchannel_producers[sc]))
        # our underlying Connection uses streaming==True, so to make things
        # easier, use an adapter when the Subchannel asks for streaming=False
        if not streaming:
            def unregister():
                self.subchannel_unregisterProducer(sc)
            producer = PullToPush(producer, unregister, self._cooperator)

        self._subchannel_producers[sc] = producer
        self._all_producers.append(producer)
        if self._paused:
            self._paused_producers.add(producer)
        else:
            self._unpaused_producers.add(producer)
        self._check_invariants()
        if streaming:
            if self._paused:
                # IPushProducers need to be paused immediately, before they
                # speak
                producer.pauseProducing()  # you wake up sleeping
        else:
            # our PullToPush adapter must be started, but if we're paused then
            # we tell it to pause before it gets a chance to write anything
            producer.startStreaming(self._paused)

    def subchannel_unregisterProducer(self, sc):
        # TODO: what if the subchannel closes, so we unregister their
        # producer for them, then the application reacts to connectionLost
        # with a duplicate unregisterProducer?
        p = self._subchannel_producers.pop(sc)
        if isinstance(p, PullToPush):
            p.stopStreaming()
        self._all_producers.remove(p)
        self._paused_producers.discard(p)
        self._unpaused_producers.discard(p)
        self._check_invariants()

    def subchannel_closed(self, scid, sc):
        self._check_invariants()
        if sc in self._subchannel_producers:
            self.subchannel_unregisterProducer(sc)

    # our Manager tells us when we've got a new Connection to work with

    def use_connection(self, c):
        self._connection = c
        assert not self._queued_unsent
        self._queued_unsent.extend(self._outbound_queue)
        # the connection can tell us to pause when we send too much data
        c.transport.registerProducer(self, True)  # IPushProducer: pause+resume
        # send our queued messages
        self.resumeProducing()

    def stop_using_connection(self):
        self._connection.transport.unregisterProducer()
        self._connection = None
        self._queued_unsent.clear()
        self.pauseProducing()
        # TODO: I expect this will call pauseProducing twice: the first time
        # when we get stopProducing (since we're registere with the
        # underlying connection as the producer), and again when the manager
        # notices the connectionLost and calls our _stop_using_connection

    def handle_ack(self, resp_seqnum):
        # we've received an inbound ack, so retire something
        while (self._outbound_queue and
               self._outbound_queue[0].seqnum <= resp_seqnum):
            self._outbound_queue.popleft()
        while (self._queued_unsent and
               self._queued_unsent[0].seqnum <= resp_seqnum):
            self._queued_unsent.popleft()
        # Inbound is responsible for tracking the high watermark and deciding
        # whether to ignore inbound messages or not

    # IPushProducer: the active connection calls these because we used
    # c.transport.registerProducer to ask for them

    def pauseProducing(self):
        if self._paused:
            return  # someone is confused and called us twice
        self._paused = True
        for p in self._all_producers:
            if p in self._unpaused_producers:
                self._unpaused_producers.remove(p)
                self._paused_producers.add(p)
                p.pauseProducing()

    def resumeProducing(self):
        if not self._paused:
            return  # someone is confused and called us twice
        self._paused = False

        while not self._paused:
            if self._queued_unsent:
                r = self._queued_unsent.popleft()
                self._connection.send_record(r)
                continue
            p = self._get_next_unpaused_producer()
            if not p:
                break
            self._paused_producers.remove(p)
            self._unpaused_producers.add(p)
            p.resumeProducing()

    def _get_next_unpaused_producer(self):
        self._check_invariants()
        if not self._paused_producers:
            return None
        while True:
            p = self._all_producers[0]
            self._all_producers.rotate(-1)  # p moves to the end of the line
            # the only unpaused Producers are at the end of the list
            assert p in self._paused_producers
            return p

    def stopProducing(self):
        # we'll hopefully have a new connection to work with in the future,
        # so we don't shut anything down. We do pause everyone, though.
        self.pauseProducing()


# modelled after twisted.internet._producer_helper._PullToPush , but with a
# configurable Cooperator, a pause-immediately argument to startStreaming()
@implementer(IPushProducer)
@attrs(eq=False)
class PullToPush(object):
    _producer = attrib(validator=provides(IPullProducer))
    _unregister = attrib(validator=lambda _a, _b, v: callable(v))
    _cooperator = attrib()
    _finished = False

    def _pull(self):
        while True:
            try:
                self._producer.resumeProducing()
            except Exception:
                log.err(None, "%s failed, producing will be stopped:" %
                        (safe_str(self._producer),))
                try:
                    self._unregister()
                    # The consumer should now call stopStreaming() on us,
                    # thus stopping the streaming.
                except Exception:
                    # Since the consumer blew up, we may not have had
                    # stopStreaming() called, so we just stop on our own:
                    log.err(None, "%s failed to unregister producer:" %
                            (safe_str(self._unregister),))
                    self._finished = True
                    return
            yield None

    def startStreaming(self, paused):
        self._coopTask = self._cooperator.cooperate(self._pull())
        if paused:
            self.pauseProducing()  # timer is scheduled, but task is removed

    def stopStreaming(self):
        if self._finished:
            return
        self._finished = True
        self._coopTask.stop()

    def pauseProducing(self):
        self._coopTask.pause()

    def resumeProducing(self):
        self._coopTask.resume()

    def stopProducing(self):
        self.stopStreaming()
        self._producer.stopProducing()



from __future__ import absolute_import, print_function, unicode_literals
if __name__ == "__main__":
    from .cli import cli
    cli.wormhole()
else:
    # raise ImportError('this module should not be imported')
    pass



import os, sys, json, contextlib, random
from twisted.internet import task, defer, endpoints
from twisted.application import service, internet
from twisted.web import server, static, resource
from wormhole import journal, wormhole

# considerations for state management:
# * be somewhat principled about the data (e.g. have a schema)
# * discourage accidental schema changes
# * avoid surprise mutations by app code (don't hand out mutables)
# * discourage app from keeping state itself: make state object easy enough
#   to use for everything. App should only hold objects that are active
#   (Services, subscribers, etc). App must wire up these objects each time.


def parse(args):
    raise NotImplementedError


def update_my_state():
    raise NotImplementedError


class State(object):
    @classmethod
    def create_empty(klass):
        self = klass()
        # to avoid being tripped up by state-mutation side-effect bugs, we
        # hold the serialized state in RAM, and re-deserialize it each time
        # someone asks for a piece of it.    # iid->invitation_stat
        empty = {"version": 1,
                 "invitations": {},
                 "contacts": [],
                 }
        self._bytes = json.dumps(empty).encode("utf-8")
        return self

    @classmethod
    def from_filename(klass, fn):
        self = klass()
        with open(fn, "rb") as f:
            bytes = f.read()
        self._bytes = bytes
        # version check
        data = self._as_data()
        assert data["version"] == 1
        # schema check?
        return self

    def save_to_filename(self, fn):
        tmpfn = fn + ".tmp"
        with open(tmpfn, "wb") as f:
            f.write(self._bytes)
        os.rename(tmpfn, fn)

    def _as_data(self):
        return json.loads(bytes.decode("utf-8"))

    @contextlib.contextmanager
    def _mutate(self):
        data = self._as_data()
        yield data  # mutable
        self._bytes = json.dumps(data).encode("utf-8")

    def get_all_invitations(self):
        return self._as_data()["invitations"]

    def add_invitation(self, iid, invitation_state):
        with self._mutate() as data:
            data["invitations"][iid] = invitation_state

    def update_invitation(self, iid, invitation_state):
        with self._mutate() as data:
            assert iid in data["invitations"]
            data["invitations"][iid] = invitation_state

    def remove_invitation(self, iid):
        with self._mutate() as data:
            del data["invitations"][iid]

    def add_contact(self, contact):
        with self._mutate() as data:
            data["contacts"].append(contact)


class Root(resource.Resource):
    pass


class Status(resource.Resource):
    def __init__(self, c):
        resource.Resource.__init__(self)
        self._call = c

    def render_GET(self, req):
        data = self._call()
        req.setHeader(b"content-type", "text/plain")
        return data


class Action(resource.Resource):
    def __init__(self, c):
        resource.Resource.__init__(self)
        self._call = c

    def render_POST(self, req):
        req.setHeader(b"content-type", "text/plain")
        try:
            args = json.load(req.content)
        except ValueError:
            req.setResponseCode(500)
            return b"bad JSON"
        data = self._call(args)
        return data


class Agent(service.MultiService):
    def __init__(self, basedir, reactor):
        service.MultiService.__init__(self)
        self._basedir = basedir
        self._reactor = reactor

        root = Root()
        site = server.Site(root)
        ep = endpoints.serverFromString(reactor, "tcp:8220")
        internet.StreamServerEndpointService(ep, site).setServiceParent(self)

        self._jm = journal.JournalManager(self._save_state)

        root.putChild(b"", static.Data("root", "text/plain"))
        root.putChild(b"list-invitations", Status(self._list_invitations))
        root.putChild(b"invite", Action(self._invite))  # {petname:}
        root.putChild(b"accept", Action(self._accept))  # {petname:, code:}

        self._state_fn = os.path.join(self._basedir, "state.json")
        self._state = State.from_filename(self._state_fn)

        self._wormholes = {}
        for iid, invitation_state in self._state.get_all_invitations().items():
            def _dispatch(event, *args, **kwargs):
                self._dispatch_wormhole_event(iid, event, *args, **kwargs)
            w = wormhole.journaled_from_data(invitation_state["wormhole"],
                                             reactor=self._reactor,
                                             journal=self._jm,
                                             event_handler=self,
                                             event_handler_args=(iid,))
            self._wormholes[iid] = w
            w.setServiceParent(self)

    def _save_state(self):
        self._state.save_to_filename(self._state_fn)

    def _list_invitations(self):
        inv = self._state.get_all_invitations()
        lines = ["%d: %s" % (iid, inv[iid]) for iid in sorted(inv)]
        return b"\n".join(lines) + b"\n"

    def _invite(self, args):
        print("invite", args)
        petname = args["petname"]
        # it'd be better to use a unique object for the event_handler
        # correlation, but we can't store them into the state database. I'm
        # not 100% sure we need one for the database: maybe it should hold a
        # list instead, and assign lookup keys at runtime. If they really
        # need to be serializable, they should be allocated rather than
        # random.
        iid = random.randint(1, 1000)
        my_pubkey = random.randint(1, 1000)
        with self._jm.process():
            w = wormhole.journaled(reactor=self._reactor, journal=self._jm,
                                   event_handler=self,
                                   event_handler_args=(iid,))
            self._wormholes[iid] = w
            w.setServiceParent(self)
            w.get_code()  # event_handler means code returns via callback
            invitation_state = {"wormhole": w.to_data(),
                                "petname": petname,
                                "my_pubkey": my_pubkey,
                                }
            self._state.add_invitation(iid, invitation_state)
        return b"ok"

    def _accept(self, args):
        print("accept", args)
        petname = args["petname"]
        code = args["code"]
        iid = random.randint(1, 1000)
        my_pubkey = random.randint(2, 2000)
        with self._jm.process():
            w = wormhole.journaled(reactor=self._reactor, journal=self._jm,
                                   event_dispatcher=self,
                                   event_dispatcher_args=(iid,))
            w.set_code(code)
            md = {"my_pubkey": my_pubkey}
            w.send(json.dumps(md).encode("utf-8"))
            invitation_state = {"wormhole": w.to_data(),
                                "petname": petname,
                                "my_pubkey": my_pubkey,
                                }
            self._state.add_invitation(iid, invitation_state)
        return b"ok"

    # dispatch options:
    # * register one function, which takes (eventname, *args)
    #   * to handle multiple wormholes, app must give is a closure
    # * register multiple functions (one per event type)
    # * register an object, with well-known method names
    # * extra: register args and/or kwargs with the callback
    #
    # events to dispatch:
    #  generated_code(code)
    #  got_verifier(verifier_bytes)
    #  verified()
    #  got_data(data_bytes)
    #  closed()

    def wormhole_dispatch_got_code(self, code, iid):
        # we're already in a jm.process() context
        invitation_state = self._state.get_all_invitations()[iid]
        invitation_state["code"] = code
        self._state.update_invitation(iid, invitation_state)
        self._wormholes[iid].set_code(code)
        # notify UI subscribers to update the display

    def wormhole_dispatch_got_verifier(self, verifier, iid):
        pass

    def wormhole_dispatch_verified(self, _, iid):
        pass

    def wormhole_dispatch_got_data(self, data, iid):
        invitation_state = self._state.get_all_invitations()[iid]
        md = json.loads(data.decode("utf-8"))
        contact = {"petname": invitation_state["petname"],
                   "my_pubkey": invitation_state["my_pubkey"],
                   "their_pubkey": md["my_pubkey"],
                   }
        self._state.add_contact(contact)
        self._wormholes[iid].close()  # now waiting for "closed"

    def wormhole_dispatch_closed(self, _, iid):
        self._wormholes[iid].disownServiceParent()
        del self._wormholes[iid]
        self._state.remove_invitation(iid)

    def handle_app_event(self, args, ack_f):  # sample function
        # Imagine here that the app has received a message (not
        # wormhole-related) from some other server, and needs to act on it.
        # Also imagine that ack_f() is how we tell the sender that they can
        # stop sending the message, or how we ask our poller/subscriber
        # client to send a DELETE message. If the process dies before ack_f()
        # delivers whatever it needs to deliver, then in the next launch,
        # handle_app_event() will be called again.
        stuff = parse(args)  # noqa
        with self._jm.process():
            update_my_state()
            self._jm.queue_outbound(ack_f)


def create(reactor, basedir):
    os.mkdir(basedir)
    s = State.create_empty()
    s.save(os.path.join(basedir, "state.json"))
    return defer.succeed(None)


def run(reactor, basedir):
    a = Agent(basedir, reactor)
    a.startService()
    print("agent listening on http://localhost:8220/")
    d = defer.Deferred()
    return d


if __name__ == "__main__":
    command = sys.argv[1]
    basedir = sys.argv[2]
    if command == "create":
        task.react(create, (basedir,))
    elif command == "run":
        task.react(run, (basedir,))
    else:
        print("Unrecognized subcommand '%s'" % command)
        sys.exit(1)



from __future__ import print_function, unicode_literals

import io
import re

from twisted.internet import reactor
from twisted.internet.defer import gatherResults, inlineCallbacks, returnValue
from twisted.internet.error import ConnectionRefusedError
from twisted.trial import unittest

from unittest import mock

from .. import _rendezvous, wormhole
from ..errors import (KeyFormatError, LonelyError, NoKeyError,
                      OnlyOneCodeError, ServerConnectionError, WormholeClosed,
                      WrongPasswordError)
from ..eventual import EventualQueue
from ..transit import allocate_tcp_port
from .common import ServerBase, poll_until

APPID = "appid"

# event orderings to exercise:
#
# * normal sender: set_code, send_phase1, connected, claimed, learn_msg2,
#   learn_phase1
# * normal receiver (argv[2]=code): set_code, connected, learn_msg1,
#   learn_phase1, send_phase1,
# * normal receiver (readline): connected, input_code
# *
# * set_code, then connected
# * connected, receive_pake, send_phase, set_code


class Delegate:
    def __init__(self):
        self.welcome = None
        self.code = None
        self.key = None
        self.verifier = None
        self.versions = None
        self.messages = []
        self.closed = None

    def wormhole_got_welcome(self, welcome):
        self.welcome = welcome

    def wormhole_got_code(self, code):
        self.code = code

    def wormhole_got_unverified_key(self, key):
        self.key = key

    def wormhole_got_verifier(self, verifier):
        self.verifier = verifier

    def wormhole_got_versions(self, versions):
        self.versions = versions

    def wormhole_got_message(self, data):
        self.messages.append(data)

    def wormhole_closed(self, result):
        self.closed = result


class Delegated(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def test_delegated(self):
        dg = Delegate()
        w1 = wormhole.create(APPID, self.relayurl, reactor, delegate=dg)
        # w1.debug_set_trace("W1")
        with self.assertRaises(NoKeyError):
            w1.derive_key("purpose", 12)
        w1.set_code("1-abc")
        self.assertEqual(dg.code, "1-abc")
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w2.set_code(dg.code)
        yield poll_until(lambda: dg.key is not None)
        yield poll_until(lambda: dg.verifier is not None)
        yield poll_until(lambda: dg.versions is not None)

        w1.send_message(b"ping")
        got = yield w2.get_message()
        self.assertEqual(got, b"ping")
        w2.send_message(b"pong")
        yield poll_until(lambda: dg.messages)
        self.assertEqual(dg.messages[0], b"pong")

        key1 = w1.derive_key("purpose", 16)
        self.assertEqual(len(key1), 16)
        self.assertEqual(type(key1), type(b""))
        with self.assertRaises(TypeError):
            w1.derive_key(b"not unicode", 16)
        with self.assertRaises(TypeError):
            w1.derive_key(12345, 16)

        w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_allocate_code(self):
        dg = Delegate()
        w1 = wormhole.create(APPID, self.relayurl, reactor, delegate=dg)
        w1.allocate_code()
        yield poll_until(lambda: dg.code is not None)
        w1.close()

    @inlineCallbacks
    def test_input_code(self):
        dg = Delegate()
        w1 = wormhole.create(APPID, self.relayurl, reactor, delegate=dg)
        h = w1.input_code()
        h.choose_nameplate("123")
        h.choose_words("purple-elephant")
        yield poll_until(lambda: dg.code is not None)
        w1.close()


class Wormholes(ServerBase, unittest.TestCase):
    # integration test, with a real server

    @inlineCallbacks
    def setUp(self):
        # test_welcome wants to see [current_cli_version]
        yield self._setup_relay(None, advertise_version="advertised.version")

    def doBoth(self, d1, d2):
        return gatherResults([d1, d2], True)

    @inlineCallbacks
    def test_allocate_default(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w1.allocate_code()
        code = yield w1.get_code()
        mo = re.search(r"^\d+-\w+-\w+$", code)
        self.assert_(mo, code)
        # w.close() fails because we closed before connecting
        yield self.assertFailure(w1.close(), LonelyError)

    @inlineCallbacks
    def test_allocate_more_words(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w1.allocate_code(3)
        code = yield w1.get_code()
        mo = re.search(r"^\d+-\w+-\w+-\w+$", code)
        self.assert_(mo, code)
        yield self.assertFailure(w1.close(), LonelyError)

    @inlineCallbacks
    def test_basic(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        # w1.debug_set_trace("W1")
        with self.assertRaises(NoKeyError):
            w1.derive_key("purpose", 12)

        w2 = wormhole.create(APPID, self.relayurl, reactor)
        # w2.debug_set_trace("  W2")
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)

        yield w1.get_unverified_key()
        yield w2.get_unverified_key()

        key1 = w1.derive_key("purpose", 16)
        self.assertEqual(len(key1), 16)
        self.assertEqual(type(key1), type(b""))
        with self.assertRaises(TypeError):
            w1.derive_key(b"not unicode", 16)
        with self.assertRaises(TypeError):
            w1.derive_key(12345, 16)

        verifier1 = yield w1.get_verifier()
        verifier2 = yield w2.get_verifier()
        self.assertEqual(verifier1, verifier2)

        versions1 = yield w1.get_versions()
        versions2 = yield w2.get_versions()
        # app-versions are exercised properly in test_versions, this just
        # tests the defaults
        self.assertEqual(versions1, {})
        self.assertEqual(versions2, {})

        w1.send_message(b"data1")
        w2.send_message(b"data2")
        dataX = yield w1.get_message()
        dataY = yield w2.get_message()
        self.assertEqual(dataX, b"data2")
        self.assertEqual(dataY, b"data1")

        versions1_again = yield w1.get_versions()
        self.assertEqual(versions1, versions1_again)

        c1 = yield w1.close()
        self.assertEqual(c1, "happy")
        c2 = yield w2.close()
        self.assertEqual(c2, "happy")

    @inlineCallbacks
    def test_get_code_early(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _eventual_queue=eq)
        d = w1.get_code()
        w1.set_code("1-abc")
        yield eq.flush()
        code = self.successResultOf(d)
        self.assertEqual(code, "1-abc")
        yield self.assertFailure(w1.close(), LonelyError)

    @inlineCallbacks
    def test_get_code_late(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _eventual_queue=eq)
        w1.set_code("1-abc")
        d = w1.get_code()
        yield eq.flush()
        code = self.successResultOf(d)
        self.assertEqual(code, "1-abc")
        yield self.assertFailure(w1.close(), LonelyError)

    @inlineCallbacks
    def test_same_message(self):
        # the two sides use random nonces for their messages, so it's ok for
        # both to try and send the same body: they'll result in distinct
        # encrypted messages
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)
        w1.send_message(b"data")
        w2.send_message(b"data")
        dataX = yield w1.get_message()
        dataY = yield w2.get_message()
        self.assertEqual(dataX, b"data")
        self.assertEqual(dataY, b"data")
        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_interleaved(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)
        w1.send_message(b"data1")
        dataY = yield w2.get_message()
        self.assertEqual(dataY, b"data1")
        d = w1.get_message()
        w2.send_message(b"data2")
        dataX = yield d
        self.assertEqual(dataX, b"data2")
        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_unidirectional(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)
        w1.send_message(b"data1")
        dataY = yield w2.get_message()
        self.assertEqual(dataY, b"data1")
        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_early(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w1.send_message(b"data1")
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        d = w2.get_message()
        w1.set_code("123-abc-def")
        w2.set_code("123-abc-def")
        dataY = yield d
        self.assertEqual(dataY, b"data1")
        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_fixed_code(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w1.set_code("123-purple-elephant")
        w2.set_code("123-purple-elephant")
        w1.send_message(b"data1"), w2.send_message(b"data2")
        dl = yield self.doBoth(w1.get_message(), w2.get_message())
        (dataX, dataY) = dl
        self.assertEqual(dataX, b"data2")
        self.assertEqual(dataY, b"data1")
        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_input_code(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w1.set_code("123-purple-elephant")
        h = w2.input_code()
        h.choose_nameplate("123")
        # Pause to allow some messages to get delivered. Specifically we want
        # to wait until w2 claims the nameplate, opens the mailbox, and
        # receives the PAKE message, to exercise the PAKE-before-CODE path in
        # Key.
        yield poll_until(lambda: w2._boss._K._debug_pake_stashed)
        h.choose_words("purple-elephant")

        w1.send_message(b"data1"), w2.send_message(b"data2")
        dl = yield self.doBoth(w1.get_message(), w2.get_message())
        (dataX, dataY) = dl
        self.assertEqual(dataX, b"data2")
        self.assertEqual(dataY, b"data1")
        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_multiple_messages(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w1.set_code("123-purple-elephant")
        w2.set_code("123-purple-elephant")
        w1.send_message(b"data1"), w2.send_message(b"data2")
        w1.send_message(b"data3"), w2.send_message(b"data4")
        dl = yield self.doBoth(w1.get_message(), w2.get_message())
        (dataX, dataY) = dl
        self.assertEqual(dataX, b"data2")
        self.assertEqual(dataY, b"data1")
        dl = yield self.doBoth(w1.get_message(), w2.get_message())
        (dataX, dataY) = dl
        self.assertEqual(dataX, b"data4")
        self.assertEqual(dataY, b"data3")
        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_closed(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _eventual_queue=eq)
        w2 = wormhole.create(APPID, self.relayurl, reactor, _eventual_queue=eq)
        w1.set_code("123-foo")
        w2.set_code("123-foo")

        # let it connect and become HAPPY
        yield w1.get_versions()
        yield w2.get_versions()

        yield w1.close()
        yield w2.close()

        # once closed, all Deferred-yielding API calls get an prompt error
        yield self.assertFailure(w1.get_welcome(), WormholeClosed)
        e = yield self.assertFailure(w1.get_code(), WormholeClosed)
        self.assertEqual(e.args[0], "happy")
        yield self.assertFailure(w1.get_unverified_key(), WormholeClosed)
        yield self.assertFailure(w1.get_verifier(), WormholeClosed)
        yield self.assertFailure(w1.get_versions(), WormholeClosed)
        yield self.assertFailure(w1.get_message(), WormholeClosed)

    @inlineCallbacks
    def test_closed_idle(self):
        yield self._relay_server.disownServiceParent()
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        # without a relay server, this won't ever connect

        d_welcome = w1.get_welcome()
        self.assertNoResult(d_welcome)
        d_code = w1.get_code()
        d_key = w1.get_unverified_key()
        d_verifier = w1.get_verifier()
        d_versions = w1.get_versions()
        d_message = w1.get_message()

        yield self.assertFailure(w1.close(), LonelyError)

        yield self.assertFailure(d_welcome, LonelyError)
        yield self.assertFailure(d_code, LonelyError)
        yield self.assertFailure(d_key, LonelyError)
        yield self.assertFailure(d_verifier, LonelyError)
        yield self.assertFailure(d_versions, LonelyError)
        yield self.assertFailure(d_message, LonelyError)

    @inlineCallbacks
    def test_wrong_password(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _eventual_queue=eq)
        w2 = wormhole.create(APPID, self.relayurl, reactor, _eventual_queue=eq)
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code + "not")
        code2 = yield w2.get_code()
        self.assertNotEqual(code, code2)
        # That's enough to allow both sides to discover the mismatch, but
        # only after the confirmation message gets through. API calls that
        # don't wait will appear to work until the mismatched confirmation
        # message arrives.
        w1.send_message(b"should still work")
        w2.send_message(b"should still work")

        key2 = yield w2.get_unverified_key()  # should work
        # w2 has just received w1.PAKE, and is about to send w2.VERSION
        key1 = yield w1.get_unverified_key()  # should work
        # w1 has just received w2.PAKE, and is about to send w1.VERSION, and
        # then will receive w2.VERSION. When it sees w2.VERSION, it will
        # learn about the WrongPasswordError.
        self.assertNotEqual(key1, key2)

        # API calls that wait (i.e. get) will errback. We collect all these
        # Deferreds early to exercise the wait-then-fail path
        d1_verified = w1.get_verifier()
        d1_versions = w1.get_versions()
        d1_received = w1.get_message()
        d2_verified = w2.get_verifier()
        d2_versions = w2.get_versions()
        d2_received = w2.get_message()

        # wait for each side to notice the failure
        yield self.assertFailure(w1.get_verifier(), WrongPasswordError)
        yield self.assertFailure(w2.get_verifier(), WrongPasswordError)
        # the rest of the loops should fire within the next tick
        yield eq.flush()

        # now all the rest should have fired already
        self.failureResultOf(d1_verified, WrongPasswordError)
        self.failureResultOf(d1_versions, WrongPasswordError)
        self.failureResultOf(d1_received, WrongPasswordError)
        self.failureResultOf(d2_verified, WrongPasswordError)
        self.failureResultOf(d2_versions, WrongPasswordError)
        self.failureResultOf(d2_received, WrongPasswordError)

        # and at this point, with the failure safely noticed by both sides,
        # new get_unverified_key() calls should signal the failure, even
        # before we close

        # any new calls in the error state should immediately fail
        yield self.assertFailure(w1.get_unverified_key(), WrongPasswordError)
        yield self.assertFailure(w1.get_verifier(), WrongPasswordError)
        yield self.assertFailure(w1.get_versions(), WrongPasswordError)
        yield self.assertFailure(w1.get_message(), WrongPasswordError)
        yield self.assertFailure(w2.get_unverified_key(), WrongPasswordError)
        yield self.assertFailure(w2.get_verifier(), WrongPasswordError)
        yield self.assertFailure(w2.get_versions(), WrongPasswordError)
        yield self.assertFailure(w2.get_message(), WrongPasswordError)

        yield self.assertFailure(w1.close(), WrongPasswordError)
        yield self.assertFailure(w2.close(), WrongPasswordError)

        # API calls should still get the error, not WormholeClosed
        yield self.assertFailure(w1.get_unverified_key(), WrongPasswordError)
        yield self.assertFailure(w1.get_verifier(), WrongPasswordError)
        yield self.assertFailure(w1.get_versions(), WrongPasswordError)
        yield self.assertFailure(w1.get_message(), WrongPasswordError)
        yield self.assertFailure(w2.get_unverified_key(), WrongPasswordError)
        yield self.assertFailure(w2.get_verifier(), WrongPasswordError)
        yield self.assertFailure(w2.get_versions(), WrongPasswordError)
        yield self.assertFailure(w2.get_message(), WrongPasswordError)

    @inlineCallbacks
    def test_wrong_password_with_spaces(self):
        w = wormhole.create(APPID, self.relayurl, reactor)
        badcode = "4 oops spaces"
        with self.assertRaises(KeyFormatError) as ex:
            w.set_code(badcode)
        expected_msg = "Code '%s' contains spaces." % (badcode, )
        self.assertEqual(expected_msg, str(ex.exception))
        yield self.assertFailure(w.close(), LonelyError)

    @inlineCallbacks
    def test_wrong_password_with_leading_space(self):
        w = wormhole.create(APPID, self.relayurl, reactor)
        badcode = " 4-oops-space"
        with self.assertRaises(KeyFormatError) as ex:
            w.set_code(badcode)
        expected_msg = "Code '%s' contains spaces." % (badcode, )
        self.assertEqual(expected_msg, str(ex.exception))
        yield self.assertFailure(w.close(), LonelyError)

    @inlineCallbacks
    def test_wrong_password_with_non_numeric_nameplate(self):
        w = wormhole.create(APPID, self.relayurl, reactor)
        badcode = "four-oops-space"
        with self.assertRaises(KeyFormatError) as ex:
            w.set_code(badcode)
        expected_msg = "Nameplate 'four' must be numeric, with no spaces."
        self.assertEqual(expected_msg, str(ex.exception))
        yield self.assertFailure(w.close(), LonelyError)

    @inlineCallbacks
    def test_welcome(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        wel1 = yield w1.get_welcome()  # early: before connection established
        wel2 = yield w1.get_welcome()  # late: already received welcome
        self.assertEqual(wel1, wel2)
        self.assertIn("current_cli_version", wel1)

        # cause an error, so a later get_welcome will return the error
        w1.set_code("123-foo")
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w2.set_code("123-NOT")
        yield self.assertFailure(w1.get_verifier(), WrongPasswordError)
        yield self.assertFailure(w1.get_welcome(), WrongPasswordError)  # late

        # we have to ensure w2 receives a "bad" message from w1 before
        # the w2.close() assertion below will actually fail
        yield self.assertFailure(w2.get_verifier(), WrongPasswordError)

        yield self.assertFailure(w1.close(), WrongPasswordError)
        yield self.assertFailure(w2.close(), WrongPasswordError)

    @inlineCallbacks
    def test_verifier(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _eventual_queue=eq)
        w2 = wormhole.create(APPID, self.relayurl, reactor, _eventual_queue=eq)
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)
        v1 = yield w1.get_verifier()  # early
        v2 = yield w2.get_verifier()
        self.failUnlessEqual(type(v1), type(b""))
        self.failUnlessEqual(v1, v2)
        w1.send_message(b"data1")
        w2.send_message(b"data2")
        dataX = yield w1.get_message()
        dataY = yield w2.get_message()
        self.assertEqual(dataX, b"data2")
        self.assertEqual(dataY, b"data1")

        # calling get_verifier() this late should fire right away
        d = w2.get_verifier()
        yield eq.flush()
        v1_late = self.successResultOf(d)
        self.assertEqual(v1_late, v1)

        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_versions(self):
        # there's no API for this yet, but make sure the internals work
        w1 = wormhole.create(
            APPID, self.relayurl, reactor, versions={"w1": 123})
        w2 = wormhole.create(
            APPID, self.relayurl, reactor, versions={"w2": 456})
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)
        w1_versions = yield w2.get_versions()
        self.assertEqual(w1_versions, {"w1": 123})
        w2_versions = yield w1.get_versions()
        self.assertEqual(w2_versions, {"w2": 456})
        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_rx_dedup(self):
        # Future clients will handle losing/reestablishing the Rendezvous
        # Server connection by retransmitting messages, which will sometimes
        # cause duplicate messages. Make sure this client can tolerate them.
        # The first place this would fail was when the second copy of the
        # incoming PAKE message was received, which would cause
        # SPAKE2.finish() to be called a second time, which throws an error
        # (which, being somewhat unexpected, caused a hang rather than a
        # clear exception). The Mailbox object is responsible for
        # deduplication, so we must patch the RendezvousConnector to simulate
        # duplicated messages.
        with mock.patch("wormhole._boss.RendezvousConnector", MessageDoubler):
            w1 = wormhole.create(APPID, self.relayurl, reactor)
        w2 = wormhole.create(APPID, self.relayurl, reactor)
        w1.set_code("123-purple-elephant")
        w2.set_code("123-purple-elephant")
        w1.send_message(b"data1"), w2.send_message(b"data2")
        dl = yield self.doBoth(w1.get_message(), w2.get_message())
        (dataX, dataY) = dl
        self.assertEqual(dataX, b"data2")
        self.assertEqual(dataY, b"data1")
        yield w1.close()
        yield w2.close()


class MessageDoubler(_rendezvous.RendezvousConnector):
    # we could double messages on the sending side, but a future server will
    # strip those duplicates, so to really exercise the receiver, we must
    # double them on the inbound side instead
    # def _msg_send(self, phase, body):
    #     wormhole._Wormhole._msg_send(self, phase, body)
    #     self._ws_send_command("add", phase=phase, body=bytes_to_hexstr(body))
    def _response_handle_message(self, msg):
        _rendezvous.RendezvousConnector._response_handle_message(self, msg)
        _rendezvous.RendezvousConnector._response_handle_message(self, msg)


class Errors(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def test_derive_key_early(self):
        w = wormhole.create(APPID, self.relayurl, reactor)
        # definitely too early
        with self.assertRaises(NoKeyError):
            w.derive_key("purpose", 12)
        yield self.assertFailure(w.close(), LonelyError)

    @inlineCallbacks
    def test_multiple_set_code(self):
        w = wormhole.create(APPID, self.relayurl, reactor)
        w.set_code("123-purple-elephant")
        # code can only be set once
        with self.assertRaises(OnlyOneCodeError):
            w.set_code("123-nope")
        yield self.assertFailure(w.close(), LonelyError)

    @inlineCallbacks
    def test_allocate_and_set_code(self):
        w = wormhole.create(APPID, self.relayurl, reactor)
        w.allocate_code()
        yield w.get_code()
        with self.assertRaises(OnlyOneCodeError):
            w.set_code("123-nope")
        yield self.assertFailure(w.close(), LonelyError)


class Reconnection(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def test_basic(self):
        w1 = wormhole.create(APPID, self.relayurl, reactor)
        w1_in = []
        w1._boss._RC._debug_record_inbound_f = w1_in.append
        # w1.debug_set_trace("W1")
        w1.allocate_code()
        code = yield w1.get_code()
        w1.send_message(b"data1")  # queued until wormhole is established

        # now wait until we've deposited all our messages on the server
        def seen_our_pake():
            for m in w1_in:
                if m["type"] == "message" and m["phase"] == "pake":
                    return True
            return False

        yield poll_until(seen_our_pake)

        w1_in[:] = []
        # drop the connection
        w1._boss._RC._ws.transport.loseConnection()
        # wait for it to reconnect and redeliver all the messages. The server
        # sends mtype=message messages in random order, but we've only sent
        # one of them, so it's safe to wait for just the PAKE phase.
        yield poll_until(seen_our_pake)

        # now let the second side proceed. this simulates the most common
        # case: the server is bounced while the sender is waiting, before the
        # receiver has started

        w2 = wormhole.create(APPID, self.relayurl, reactor)
        # w2.debug_set_trace("  W2")
        w2.set_code(code)

        dataY = yield w2.get_message()
        self.assertEqual(dataY, b"data1")

        w2.send_message(b"data2")
        dataX = yield w1.get_message()
        self.assertEqual(dataX, b"data2")

        c1 = yield w1.close()
        self.assertEqual(c1, "happy")
        c2 = yield w2.close()
        self.assertEqual(c2, "happy")


class InitialFailure(unittest.TestCase):
    @inlineCallbacks
    def assertSCEFailure(self, eq, d, innerType):
        yield eq.flush()
        f = self.failureResultOf(d, ServerConnectionError)
        inner = f.value.reason
        self.assertIsInstance(inner, innerType)
        returnValue(inner)

    @inlineCallbacks
    def test_bad_dns(self):
        eq = EventualQueue(reactor)
        # point at a URL that will never connect
        w = wormhole.create(
            APPID, "ws://%%%.example.org:4000/v1", reactor, _eventual_queue=eq)
        # that should have already received an error, when it tried to
        # resolve the bogus DNS name. All API calls will return an error.

        e = yield self.assertSCEFailure(eq, w.get_unverified_key(), ValueError)
        self.assertIsInstance(e, ValueError)
        self.assertEqual(str(e), "invalid hostname: %%%.example.org")
        yield self.assertSCEFailure(eq, w.get_code(), ValueError)
        yield self.assertSCEFailure(eq, w.get_verifier(), ValueError)
        yield self.assertSCEFailure(eq, w.get_versions(), ValueError)
        yield self.assertSCEFailure(eq, w.get_message(), ValueError)

    @inlineCallbacks
    def assertSCE(self, d, innerType):
        e = yield self.assertFailure(d, ServerConnectionError)
        inner = e.reason
        self.assertIsInstance(inner, innerType)
        returnValue(inner)

    @inlineCallbacks
    def test_no_connection(self):
        # point at a URL that will never connect
        port = allocate_tcp_port()
        w = wormhole.create(APPID, "ws://127.0.0.1:%d/v1" % port, reactor)
        # nothing is listening, but it will take a turn to discover that
        d1 = w.get_code()
        d2 = w.get_unverified_key()
        d3 = w.get_verifier()
        d4 = w.get_versions()
        d5 = w.get_message()
        yield self.assertSCE(d1, ConnectionRefusedError)
        yield self.assertSCE(d2, ConnectionRefusedError)
        yield self.assertSCE(d3, ConnectionRefusedError)
        yield self.assertSCE(d4, ConnectionRefusedError)
        yield self.assertSCE(d5, ConnectionRefusedError)

    @inlineCallbacks
    def test_all_deferreds(self):
        # point at a URL that will never connect
        port = allocate_tcp_port()
        w = wormhole.create(APPID, "ws://127.0.0.1:%d/v1" % port, reactor)
        # nothing is listening, but it will take a turn to discover that
        w.allocate_code()
        d1 = w.get_code()
        d2 = w.get_unverified_key()
        d3 = w.get_verifier()
        d4 = w.get_versions()
        d5 = w.get_message()
        yield self.assertSCE(d1, ConnectionRefusedError)
        yield self.assertSCE(d2, ConnectionRefusedError)
        yield self.assertSCE(d3, ConnectionRefusedError)
        yield self.assertSCE(d4, ConnectionRefusedError)
        yield self.assertSCE(d5, ConnectionRefusedError)


class Trace(unittest.TestCase):
    def test_basic(self):
        w1 = wormhole.create(APPID, "ws://localhost:1", reactor)
        stderr = io.StringIO()
        w1.debug_set_trace("W1", file=stderr)
        # if Automat doesn't have the tracing API, then we won't actually
        # exercise the tracing function, so exercise the RendezvousConnector
        # function manually (it isn't a state machine, so it will always wire
        # up the tracer)
        w1._boss._RC._debug("what")

        stderr = io.StringIO()
        out = w1._boss._print_trace("OLD", "IN", "NEW", "C1", "M1", stderr)
        self.assertEqual(stderr.getvalue().splitlines(),
                         ["C1.M1[OLD].IN -> [NEW]"])
        out("OUT1")
        self.assertEqual(stderr.getvalue().splitlines(),
                         ["C1.M1[OLD].IN -> [NEW]", " C1.M1.OUT1()"])
        w1._boss._print_trace("", "R.connected", "", "C1", "RC1", stderr)
        self.assertEqual(
            stderr.getvalue().splitlines(),
            ["C1.M1[OLD].IN -> [NEW]", " C1.M1.OUT1()", "C1.RC1.R.connected"])

    def test_delegated(self):
        dg = Delegate()
        w1 = wormhole.create(APPID, "ws://localhost:1", reactor, delegate=dg)
        stderr = io.StringIO()
        w1.debug_set_trace("W1", file=stderr)
        w1._boss._RC._debug("what")
